#  Copyright (c) 2021-2023, Manfred Moitzi
#  License: MIT License
from __future__ import annotations
from typing import Optional, TYPE_CHECKING, cast
import abc
import enum
import logging
import os
import pathlib

from ezdxf import options
from .font_face import FontFace
from .font_manager import (
    FontManager,
    SUPPORTED_TTF_TYPES,
    FontNotFoundError,
    UnsupportedFont,
)
from .font_synonyms import FONT_SYNONYMS
from .font_measurements import FontMeasurements
from .glyphs import GlyphPath, Glyphs

if TYPE_CHECKING:
    from ezdxf.document import Drawing
    from ezdxf.entities import DXFEntity, Textstyle

logger = logging.getLogger("ezdxf")
FONT_MANAGER_CACHE_FILE = "font_manager_cache.json"
# SUT = System Under Test, see build_sut_font_manager_cache() function
SUT_FONT_MANAGER_CACHE = False
CACHE_DIRECTORY = ".cache"
font_manager = FontManager()

SHX_FONTS = {
    # See examples in: CADKitSamples/Shapefont.dxf
    # Shape file structure is not documented, therefore, replace these fonts by
    # true type fonts.
    # `None` is for: use the default font.
    #
    # All these replacement TTF fonts have a copyright remark:
    # "(c) Copyright 1996 by Autodesk Inc., All rights reserved"
    # and therefore can not be included in ezdxf or the associated repository!
    # You got them if you install any Autodesk product, like the free available
    # DWG/DXF viewer "TrueView" : https://www.autodesk.com/viewers
    "AMGDT": "amgdt___.ttf",  # Tolerance symbols
    "AMGDT.SHX": "amgdt___.ttf",
    "COMPLEX": "complex_.ttf",
    "COMPLEX.SHX": "complex_.ttf",
    "ISOCP": "isocp.ttf",
    "ISOCP.SHX": "isocp.ttf",
    "ITALIC": "italicc_.ttf",
    "ITALIC.SHX": "italicc_.ttf",
    "GOTHICG": "gothicg_.ttf",
    "GOTHICG.SHX": "gothicg_.ttf",
    "GREEKC": "greekc.ttf",
    "GREEKC.SHX": "greekc.ttf",
    "ROMANS": "romans__.ttf",
    "ROMANS.SHX": "romans__.ttf",
    "SCRIPTS": "scripts_.ttf",
    "SCRIPTS.SHX": "scripts_.ttf",
    "SCRIPTC": "scriptc_.ttf",
    "SCRIPTC.SHX": "scriptc_.ttf",
    "SIMPLEX": "simplex_.ttf",
    "SIMPLEX.SHX": "simplex_.ttf",
    "SYMATH": "symath__.ttf",
    "SYMATH.SHX": "symath__.ttf",
    "SYMAP": "symap___.ttf",
    "SYMAP.SHX": "symap___.ttf",
    "SYMETEO": "symeteo_.ttf",
    "SYMETEO.SHX": "symeteo_.ttf",
    "TXT": "txt_____.ttf",  # Default AutoCAD font
    "TXT.SHX": "txt_____.ttf",
}
LFF_FONTS = {
    "TXT": "standard.lff",
    "TXT.SHX": "standard.lff",
}
TTF_TO_SHX = {v: k for k, v in SHX_FONTS.items() if k.endswith("SHX")}
DESCENDER_FACTOR = 0.333  # from TXT SHX font - just guessing
X_HEIGHT_FACTOR = 0.666  # from TXT SHX font - just guessing
MONOSPACE = "*monospace"  # last resort fallback font only for measurement


def map_shx_to_ttf(font_name: str) -> str:
    """Map .shx font names to .ttf file names. e.g. "TXT" -> "txt_____.ttf" """
    # Map SHX fonts to True Type Fonts:
    font_upper = font_name.upper()
    if font_upper in SHX_FONTS:
        font_name = SHX_FONTS[font_upper]
    return font_name


def map_shx_to_lff(font_name: str) -> str:
    """Map .shx font names to .lff file names. e.g. "TXT" -> "standard.lff" """
    font_upper = font_name.upper()
    name = LFF_FONTS.get(font_upper, "")
    if font_manager.has_font(name):
        return name
    if not font_upper.endswith(".SHX"):
        lff_name = font_name + ".lff"
    else:
        lff_name = font_name[:-4] + ".lff"
    if font_manager.has_font(lff_name):
        return lff_name
    return font_name


def is_shx_font_name(font_name: str) -> bool:
    name = font_name.lower()
    if name.endswith(".shx"):
        return True
    if "." not in name:
        return True
    return False


def map_ttf_to_shx(ttf: str) -> Optional[str]:
    """Maps .ttf filenames to .shx font names. e.g. "txt_____.ttf" -> "TXT" """
    return TTF_TO_SHX.get(ttf.lower())


def build_system_font_cache() -> None:
    """Builds or rebuilds the font manager cache. The font manager cache has a fixed
    location in the cache directory of the users home directory "~/.cache/ezdxf" or the
    directory specified by the environment variable "XDG_CACHE_HOME".
    """
    build_font_manager_cache(_get_font_manager_path())


def find_font_face(font_name: str) -> FontFace:
    """Returns the :class:`FontFace` definition for the given font filename
    e.g. "LiberationSans-Regular.ttf".

    """
    return font_manager.get_font_face(font_name)


def get_font_face(font_name: str, map_shx=True) -> FontFace:
    """Returns  the :class:`FontFace` definition for the given font filename
    e.g. "LiberationSans-Regular.ttf".

    This function translates a DXF font definition by the TTF font file name into a
    :class:`FontFace` object. Returns the :class:`FontFace` of the default font when a
    font is not available on the current system.

    Args:
        font_name: raw font file name as stored in the
            :class:`~ezdxf.entities.Textstyle` entity
        map_shx: maps SHX font names to TTF replacement fonts,
            e.g. "TXT" -> "txt_____.ttf"

    """
    if not isinstance(font_name, str):
        raise TypeError("font_name has invalid type")
    if map_shx:
        font_name = map_shx_to_ttf(font_name)
    return find_font_face(font_name)


def resolve_shx_font_name(font_name: str, order: str) -> str:
    """Resolves a .shx font name, the argument `order` defines the resolve order:

    - "t" = map .shx fonts to TrueType fonts (.ttf, .ttc, .otf)
    - "s" = use shapefile fonts (.shx, .shp)
    - "l" = map .shx fonts to LibreCAD fonts (.lff)

    """
    if len(order) == 0:
        return font_name
    order = order.lower()
    for type_str in order:
        if type_str == "t":
            name = map_shx_to_ttf(font_name)
            if font_manager.has_font(name):
                return name
        elif type_str == "s":
            if not font_name.lower().endswith(".shx"):
                font_name += ".shx"
            if font_manager.has_font(font_name):
                return font_name
        elif type_str == "l":
            name = map_shx_to_lff(font_name)
            if font_manager.has_font(name):
                return name
    return font_name


def resolve_font_face(font_name: str, order="tsl") -> FontFace:
    """Returns the :class:`FontFace` definition for the given font filename
    e.g. "LiberationSans-Regular.ttf".

    This function translates a DXF font definition by the TTF font file name into a
    :class:`FontFace` object. Returns the :class:`FontFace` of the default font when a
    font is not available on the current system.

    The order argument defines the resolve order for .shx fonts:

    - "t" = map .shx fonts to TrueType fonts (.ttf, .ttc, .otf)
    - "s" = use shapefile fonts (.shx, .shp)
    - "l" = map .shx fonts to LibreCAD fonts (.lff)

    Args:
        font_name: raw font file name as stored in the
            :class:`~ezdxf.entities.Textstyle` entity
        order: resolving order

    """
    if not isinstance(font_name, str):
        raise TypeError("font_name has invalid type")
    if is_shx_font_name(font_name):
        font_name = resolve_shx_font_name(font_name, order)
    return find_font_face(font_name)


def get_font_measurements(font_name: str, map_shx=True) -> FontMeasurements:
    """Get :class:`FontMeasurements` for the given font filename
    e.g. "LiberationSans-Regular.ttf".

    Args:
        font_name: raw font file name as stored in the
            :class:`~ezdxf.entities.Textstyle` entity
        map_shx: maps SHX font names to TTF replacement fonts,
            e.g. "TXT" -> "txt_____.ttf"

    """
    if map_shx:
        font_name = map_shx_to_ttf(font_name)
    elif is_shx_font_name(font_name):
        return FontMeasurements(
            baseline=0,
            cap_height=1,
            x_height=X_HEIGHT_FACTOR,
            descender_height=DESCENDER_FACTOR,
        )
    font = TrueTypeFont(font_name, cap_height=1)
    return font.measurements


def find_best_match(
    *,
    family: str = "sans-serif",
    style: str = "Regular",
    weight: int = 400,
    width: int = 5,
    italic: Optional[bool] = False,
) -> Optional[FontFace]:
    """Returns a :class:`FontFace` that matches the given properties best. The search
    is based the descriptive properties and not on comparing glyph shapes. Returns
    ``None`` if no font was found.

    Args:
        family: font family name e.g. "sans-serif", "Liberation Sans"
        style: font style e.g. "Regular", "Italic", "Bold"
        weight: weight in the range from 1-1000 (usWeightClass)
        width: width in the range from 1-9 (usWidthClass)
        italic: ``True``, ``False`` or ``None`` to ignore this flag

    """
    return font_manager.find_best_match(family, style, weight, width, italic)


def find_font_file_name(font_face: FontFace) -> str:
    """Returns the true type font file name without parent directories e.g. "Arial.ttf"."""
    return font_manager.find_font_name(font_face)


def load():
    """Reload all cache files. The cache files are loaded automatically at the import
    of `ezdxf`.
    """
    _load_font_manager()
    # Add font name synonyms, see discussion #1002
    # Find macOS fonts on Windows/Linux and vice versa.
    font_manager.add_synonyms(FONT_SYNONYMS, reverse=True)


def _get_font_manager_path():
    cache_path = options.xdg_path("XDG_CACHE_HOME", CACHE_DIRECTORY)
    return cache_path / FONT_MANAGER_CACHE_FILE


def _load_font_manager() -> None:
    fm_path = _get_font_manager_path()
    if fm_path.exists():
        try:
            font_manager.loads(fm_path.read_text())
            return
        except IOError as e:
            logger.info(f"Error loading cache file: {str(e)}")
    build_font_manager_cache(fm_path)


def build_sut_font_manager_cache(repo_font_path: pathlib.Path) -> None:
    """Load font manger cache for system under test (sut).

    Load the fonts included in the repository folder "./fonts" to guarantee the tests
    have the same fonts available on all systems.

    This function should be called from "conftest.py".

    """
    global SUT_FONT_MANAGER_CACHE
    SUT_FONT_MANAGER_CACHE = True
    font_manager.clear()
    cache_file = repo_font_path / "font_manager_cache.json"
    if cache_file.exists():
        try:
            font_manager.loads(cache_file.read_text())
            return
        except IOError as e:
            print(f"Error loading cache file: {str(e)}")
    font_manager.build([str(repo_font_path)], support_dirs=False)
    s = font_manager.dumps()
    try:
        cache_file.write_text(s)
    except IOError as e:
        print(f"Error writing cache file: {str(e)}")


def make_cache_directory(path: pathlib.Path) -> None:
    if not path.exists():
        try:
            path.mkdir(parents=True)
        except IOError:
            pass


def build_font_manager_cache(path: pathlib.Path) -> None:
    font_manager.clear()
    font_manager.build()
    s = font_manager.dumps()
    cache_dir = path.parent
    make_cache_directory(cache_dir)
    if not cache_dir.exists():
        logger.warning(
            f"Cannot create cache home directory: '{str(cache_dir)}', cache files will "
            f"not be saved.\nSee also issue https://github.com/mozman/ezdxf/issues/923."
        )
        return
    try:
        path.write_text(s)
    except IOError as e:
        logger.warning(f"Error writing cache file: '{str(e)}'")


class FontRenderType(enum.Enum):
    # render glyphs as filled paths: TTF, OTF
    OUTLINE = enum.auto()

    # render glyphs as line strokes: SHX, SHP
    STROKE = enum.auto


class AbstractFont:
    """The `ezdxf` font abstraction for text measurement and text path rendering."""

    font_render_type = FontRenderType.STROKE
    name: str = "undefined"

    def __init__(self, measurements: FontMeasurements):
        self.measurements = measurements

    @abc.abstractmethod
    def text_width(self, text: str) -> float:
        """Returns the text width in drawing units for the given `text` string."""
        pass

    @abc.abstractmethod
    def text_width_ex(
        self, text: str, cap_height: float, width_factor: float = 1.0
    ) -> float:
        """Returns the text width in drawing units, bypasses the stored `cap_height` and
        `width_factor`.
        """
        pass

    @abc.abstractmethod
    def space_width(self) -> float:
        """Returns the width of a "space" character a.k.a. word spacing."""
        pass

    @abc.abstractmethod
    def text_path(self, text: str) -> GlyphPath:
        """Returns the 2D text path for the given text."""
        ...

    @abc.abstractmethod
    def text_path_ex(
        self, text: str, cap_height: float, width_factor: float = 1.0
    ) -> GlyphPath:
        """Returns the 2D text path for the given text, bypasses the stored `cap_height`
        and `width_factor`."""
        ...

    @abc.abstractmethod
    def text_glyph_paths(
        self, text: str, cap_height: float, width_factor: float = 1.0
    ) -> list[GlyphPath]:
        """Returns a list of 2D glyph paths for the given text, bypasses the stored
        `cap_height` and `width_factor`."""
        ...


class MonospaceFont(AbstractFont):
    """Represents a monospaced font where each letter has the same cap- and descender
    height and the same width. The given cap height and width factor are the default
    values for measurements and rendering. The extended methods can override these
    default values.

    This font exists only for generic text measurement in tests and does not render any
    glyphs!

    """

    font_render_type = FontRenderType.STROKE
    name = MONOSPACE

    def __init__(
        self,
        cap_height: float,
        width_factor: float = 1.0,
        baseline: float = 0,
        descender_factor: float = DESCENDER_FACTOR,
        x_height_factor: float = X_HEIGHT_FACTOR,
    ):
        super().__init__(
            FontMeasurements(
                baseline=baseline,
                cap_height=cap_height,
                x_height=cap_height * x_height_factor,
                descender_height=cap_height * descender_factor,
            )
        )
        self._width_factor: float = abs(width_factor)
        self._space_width = self.measurements.cap_height * self._width_factor

    def text_width(self, text: str) -> float:
        """Returns the text width in drawing units for the given `text`."""
        return self.text_width_ex(
            text, self.measurements.cap_height, self._width_factor
        )

    def text_width_ex(
        self, text: str, cap_height: float, width_factor: float = 1.0
    ) -> float:
        """Returns the text width in drawing units, bypasses the stored `cap_height` and
        `width_factor`.
        """
        return len(text) * cap_height * width_factor

    def text_path(self, text: str) -> GlyphPath:
        """Returns the rectangle text width x cap height as :class:`NumpyPath2d` instance."""
        return self.text_path_ex(text, self.measurements.cap_height, self._width_factor)

    def text_path_ex(
        self, text: str, cap_height: float, width_factor: float = 1.0
    ) -> GlyphPath:
        """Returns the rectangle text width x cap height as  :class:`NumpyPath2d`
        instance, bypasses the stored `cap_height` and `width_factor`.
        """
        from ezdxf.path import Path

        text_width = self.text_width_ex(text, cap_height, width_factor)
        p = Path((0, 0))
        p.line_to((text_width, 0))
        p.line_to((text_width, cap_height))
        p.line_to((0, cap_height))
        p.close()
        return GlyphPath(p)

    def text_glyph_paths(
        self, text: str, cap_height: float, width_factor: float = 1.0
    ) -> list[GlyphPath]:
        """Returns the same rectangle as the method :meth:`text_path_ex` in a list."""
        return [self.text_path_ex(text, cap_height, width_factor)]

    def space_width(self) -> float:
        """Returns the width of a "space" char."""
        return self._space_width


class _CachedFont(AbstractFont, abc.ABC):
    """Abstract font with caching support."""

    _glyph_caches: dict[str, Glyphs] = dict()

    def __init__(self, font_name: str, cap_height: float, width_factor: float = 1.0):
        self.name = font_name
        cache = self.create_cache(font_name)
        self.glyph_cache = cache
        self.cap_height = float(cap_height)
        self.width_factor = float(width_factor)
        scale_factor: float = cache.get_scaling_factor(self.cap_height)
        super().__init__(cache.font_measurements.scale(scale_factor))
        self._space_width: float = (
            self.glyph_cache.space_width * scale_factor * width_factor
        )

    @abc.abstractmethod
    def create_cache(self, font_name: str) -> Glyphs:
        ...

    def text_width(self, text: str) -> float:
        """Returns the text width in drawing units for the given `text` string."""
        return self.text_width_ex(text, self.cap_height, self.width_factor)

    def text_width_ex(
        self, text: str, cap_height: float, width_factor: float = 1.0
    ) -> float:
        """Returns the text width in drawing units, bypasses the stored `cap_height` and
        `width_factor`.
        """
        if not text.strip():
            return 0
        return self.glyph_cache.get_text_length(text, cap_height, width_factor)

    def text_path(self, text: str) -> GlyphPath:
        """Returns the 2D text path for the given text."""

        return self.text_path_ex(text, self.cap_height, self.width_factor)

    def text_path_ex(
        self, text: str, cap_height: float, width_factor: float = 1.0
    ) -> GlyphPath:
        """Returns the 2D text path for the given text, bypasses the stored `cap_height`
        and `width_factor`."""
        return self.glyph_cache.get_text_path(text, cap_height, width_factor)

    def text_glyph_paths(
        self, text: str, cap_height: float, width_factor: float = 1.0
    ) -> list[GlyphPath]:
        """Returns a list of 2D glyph paths for the given text, bypasses the stored
        `cap_height` and `width_factor`."""
        return self.glyph_cache.get_text_glyph_paths(text, cap_height, width_factor)

    def space_width(self) -> float:
        """Returns the width of a "space" char."""
        return self._space_width


class TrueTypeFont(_CachedFont):
    """Represents a TrueType font. Font measurement and glyph rendering is done by the
    `fontTools` package. The given cap height and width factor are the default values
    for measurements and glyph rendering. The extended methods can override these
    default values.
    """

    font_render_type = FontRenderType.OUTLINE

    def create_cache(self, ttf: str) -> Glyphs:
        from .ttfonts import TTFontRenderer

        key = pathlib.Path(ttf).name.lower()
        try:
            return self._glyph_caches[key]
        except KeyError:
            pass
        try:
            tt_font = font_manager.get_ttf_font(ttf)
            try:  # see issue #990
                cache = TTFontRenderer(tt_font)
            except Exception:
                raise UnsupportedFont
        except UnsupportedFont:
            fallback_font_name = font_manager.fallback_font_name()
            logger.info(f"replacing unsupported font '{ttf}' by '{fallback_font_name}'")
            cache = TTFontRenderer(font_manager.get_ttf_font(fallback_font_name))
        self._glyph_caches[key] = cache
        return cache


class _UnmanagedTrueTypeFont(_CachedFont):
    font_render_type = FontRenderType.OUTLINE

    def create_cache(self, ttf: str) -> Glyphs:
        from .ttfonts import TTFontRenderer
        from fontTools.ttLib import TTFont

        key = ttf.lower()
        try:
            return self._glyph_caches[key]
        except KeyError:
            pass
        cache = TTFontRenderer(TTFont(ttf, fontNumber=0))
        self._glyph_caches[key] = cache
        return cache


def sideload_ttf(font_path: str | os.PathLike, cap_height) -> AbstractFont:
    """This function bypasses the FontManager and loads the TrueType font straight from
    the file system, requires the absolute font file path e.g. "C:/Windows/Fonts/Arial.ttf".

    .. warning::

        Expert feature, use with care: no fallback font and no error handling.

    """

    return _UnmanagedTrueTypeFont(str(font_path), cap_height)


class ShapeFileFont(_CachedFont):
    """Represents a shapefile font (.shx, .shp). Font measurement and glyph rendering is
    done by the ezdxf.fonts.shapefile module. The given cap height and width factor are
    the default values for measurements and glyph rendering. The extended methods can
    override these default values.

    """

    font_render_type = FontRenderType.STROKE

    def create_cache(self, font_name: str) -> Glyphs:
        key = font_name.lower()
        try:
            return self._glyph_caches[key]
        except KeyError:
            pass
        glyph_cache = font_manager.get_shapefile_glyph_cache(font_name)
        self._glyph_caches[key] = glyph_cache
        return glyph_cache


class LibreCadFont(_CachedFont):
    """Represents a LibreCAD font (.shx, .shp). Font measurement and glyph rendering is
    done by the ezdxf.fonts.lff module. The given cap height and width factor are the
    default values for measurements and glyph rendering. The extended methods can
    override these default values.

    """

    font_render_type = FontRenderType.STROKE

    def create_cache(self, font_name: str) -> Glyphs:
        key = font_name.lower()
        try:
            return self._glyph_caches[key]
        except KeyError:
            pass
        glyph_cache = font_manager.get_lff_glyph_cache(font_name)
        self._glyph_caches[key] = glyph_cache
        return glyph_cache


def make_font(
    font_name: str, cap_height: float, width_factor: float = 1.0
) -> AbstractFont:
    r"""Returns a font abstraction based on class :class:`AbstractFont`.

    Supported font types:

    - .ttf, .ttc and .otf - TrueType fonts
    - .shx, .shp - Autodesk® shapefile fonts
    - .lff - LibreCAD font format

    The special name "\*monospace" returns the fallback font :class:`MonospaceFont` for
    testing and basic measurements.

    .. note:: The font definition files are not included in `ezdxf`.

    Args:
        font_name: font file name as stored in the :class:`~ezdxf.entities.Textstyle`
            entity e.g. "OpenSans-Regular.ttf"
        cap_height: desired cap height in drawing units.
        width_factor: horizontal text stretch factor

    """
    if font_name == MONOSPACE:
        return MonospaceFont(cap_height, width_factor)
    ext = pathlib.Path(font_name).suffix.lower()
    last_resort = MonospaceFont(cap_height, width_factor)
    if ext in SUPPORTED_TTF_TYPES:
        try:
            return TrueTypeFont(font_name, cap_height, width_factor)
        except FontNotFoundError as e:
            logger.warning(f"no default font found: {str(e)}")
            return last_resort
    elif ext == ".shx" or ext == ".shp":
        try:
            return ShapeFileFont(font_name, cap_height, width_factor)
        except FontNotFoundError:
            pass
        except UnsupportedFont:
            # change name - the font exists but is not supported
            font_name = font_manager.fallback_font_name()
    elif ext == ".lff":
        try:
            return LibreCadFont(font_name, cap_height, width_factor)
        except FontNotFoundError:
            pass
    elif ext == "":  # e.g. "TXT"
        font_face = font_manager.find_best_match(
            family=font_name, style=".shx", italic=None
        )
        if font_face is not None:
            return make_font(font_face.filename, cap_height, width_factor)
    else:
        logger.warning(f"unsupported font-name suffix: {font_name}")
        font_name = font_manager.fallback_font_name()

    # return default TrueType font
    try:
        return TrueTypeFont(font_name, cap_height, width_factor)
    except FontNotFoundError as e:
        logger.warning(f"no default font found: {str(e)}")
        return last_resort


def get_entity_font_face(entity: DXFEntity, doc: Optional[Drawing] = None) -> FontFace:
    """Returns the :class:`FontFace` defined by the associated text style.
    Returns the default font face if the `entity` does not have or support
    the DXF attribute "style". Supports the extended font information stored in
    :class:`~ezdxf.entities.Textstyle` table entries.

    Pass a DXF document as argument `doc` to resolve text styles for virtual
    entities which are not assigned to a DXF document. The argument `doc`
    always overrides the DXF document to which the `entity` is assigned to.

    """
    if entity.doc and doc is None:
        doc = entity.doc
    if doc is None:
        return FontFace()

    style_name = ""
    # This works also for entities which do not support "style",
    # where style_name = entity.dxf.get("style") would fail.
    if entity.dxf.is_supported("style"):
        style_name = entity.dxf.style

    font_face = FontFace()
    if style_name:
        style = cast("Textstyle", doc.styles.get(style_name))
        family, italic, bold = style.get_extended_font_data()
        if family:
            text_style = "Italic" if italic else "Regular"
            text_weight = 700 if bold else 400
            font_face = FontFace(family=family, style=text_style, weight=text_weight)
        else:
            ttf = style.dxf.font
            if ttf:
                font_face = get_font_face(ttf)
    return font_face


load()



# Copyright (c) 2021, Manfred Moitzi
# License: MIT License
import pytest
import math

from ezdxf.entities import Hatch
from ezdxf.render.forms import box
from ezdxf.math import Vec2, Vec3, Matrix44, arc_angle_span_deg
from ezdxf.path import make_path, have_close_control_vertices


@pytest.fixture()
def m44():
    return Matrix44.chain(
        Matrix44.z_rotate(math.pi / 2),
        Matrix44.translate(1, 2, 0),
    )


def test_polyline_path_transform_interface(m44):
    hatch = Hatch.new()
    vertices = list(box(1.0, 2.0))
    path = hatch.paths.add_polyline_path(vertices)

    hatch.transform(m44)
    chk = m44.transform_vertices(vertices)
    for v, c in zip(path.vertices, chk):
        assert c.isclose(v)


def test_edge_path_transform_interface(m44):
    hatch = Hatch.new()
    path = hatch.paths.add_edge_path()
    path.add_line((0, 0), (10, 0))
    path.add_arc((10, 5), radius=5, start_angle=270, end_angle=450, ccw=1)
    path.add_ellipse(
        (5, 10), major_axis=(5, 0), ratio=0.2, start_angle=0, end_angle=180
    )
    spline = path.add_spline(
        [(1, 1), (2, 2), (3, 3), (4, 4)], degree=3, periodic=1
    )
    # the following values do not represent a mathematically valid spline
    spline.control_points = [(1, 1), (2, 2), (3, 3), (4, 4)]
    spline.knot_values = [1, 2, 3, 4, 5, 6]
    spline.weights = [4, 3, 2, 1]
    spline.start_tangent = (10, 1)
    spline.end_tangent = (2, 20)

    chk = list(
        m44.transform_vertices(
            [
                Vec3(0, 0),
                Vec3(10, 0),
                Vec3(10, 5),
                Vec3(5, 10),
                Vec3(1, 1),
                Vec3(2, 2),
                Vec3(3, 3),
                Vec3(4, 4),
            ]
        )
    )

    hatch.transform(m44)
    line = path.edges[0]
    assert chk[0].isclose(line.start)
    assert chk[1].isclose(line.end)
    arc = path.edges[1]
    assert chk[2].isclose(arc.center)
    ellipse = path.edges[2]
    assert chk[3].isclose(ellipse.center)
    spline = path.edges[3]
    for c, v in zip(chk[4:], spline.control_points):
        assert c.isclose(v)
    for c, v in zip(chk[4:], spline.fit_points):
        assert c.isclose(v)
    assert m44.transform_direction((10, 1, 0)).isclose(spline.start_tangent)
    assert m44.transform_direction((2, 20, 0)).isclose(spline.end_tangent)


@pytest.fixture(params=["arc", "ellipse"])
def closed_edge_hatch(request):
    _hatch = Hatch.new()
    _path = _hatch.paths.add_edge_path()
    if request.param == "arc":
        _path.add_arc((0, 0), radius=1, start_angle=0, end_angle=360, ccw=1)
    elif request.param == "ellipse":
        _path.add_ellipse(
            (0, 0), major_axis=(5, 0), ratio=0.2, start_angle=0, end_angle=360
        )
    return _hatch


def test_full_circle_ellipse_edge_rotation(closed_edge_hatch):
    edge = closed_edge_hatch.paths[0].edges[0]
    assert arc_angle_span_deg(
        edge.start_angle, edge.end_angle
    ) == pytest.approx(360)

    closed_edge_hatch.transform(Matrix44.z_rotate(math.radians(30)))
    edge2 = closed_edge_hatch.paths[0].edges[0]
    assert arc_angle_span_deg(
        edge2.start_angle, edge2.end_angle
    ) == pytest.approx(360)


def test_full_circle_edge_scaling():
    _hatch = Hatch.new()
    _path = _hatch.paths.add_edge_path()
    _arc = _path.add_arc((0, 0), radius=1, start_angle=0, end_angle=360, ccw=1)
    _hatch.transform(Matrix44.scale(0.5, 0.5, 0.5))
    assert _arc.radius == pytest.approx(0.5)


def transformed_copy(entity, matrix):
    _copy = entity.copy()
    _copy.transform(matrix)
    return _copy


@pytest.mark.parametrize(
    "sx, sy, extrusion",
    [
        (-1, 1, (0, 0, -1)),
        (1, -1, (0, 0, -1)),
        (-1, -1, (0, 0, 1)),
    ],
    ids=["mirror-x", "mirror-y", "mirror-xy"],
)
@pytest.mark.parametrize("kind", ["arc", "ellipse"])
def test_ocs_mirror_transformations_of_clockwise_oriented_curves(
    sx,
    sy,
    extrusion,
    kind,
):
    hatch = Hatch()
    edge_path = hatch.paths.add_edge_path()
    if kind == "arc":
        edge_path.add_arc((7, 0), 5, start_angle=0, end_angle=180, ccw=False)
    elif kind == "ellipse":
        edge_path.add_ellipse(
            (7, 0), (5, 0), ratio=0.7, start_angle=0, end_angle=180, ccw=False
        )
    else:
        pytest.fail(f"unknown kind: {kind}")

    transformed_hatch = transformed_copy(hatch, Matrix44.scale(sx, sy, 1))

    # This tests the current implementation of OCS transformations!
    assert transformed_hatch.dxf.extrusion.isclose(extrusion)
    assert (
        transformed_hatch.paths[0].edges[0].ccw is False
    ), "ccw flag should not change"


@pytest.mark.parametrize(
    "sx, sy",
    [(-1, 1), (1, -1), (-1, -1)],
    ids=["mirror-x", "mirror-y", "mirror-xy"],
)
@pytest.mark.parametrize("kind", ["arc", "ellipse"])
def test_wcs_mirror_transformations_of_clockwise_oriented_curves(sx, sy, kind):
    hatch = Hatch()
    edge_path = hatch.paths.add_edge_path()
    # A closed loop is required to get a path!
    edge_path.add_line((15, 5), (5, 5))
    if kind == "arc":
        edge_path.add_arc((10, 5), 5, start_angle=0, end_angle=180, ccw=False)
    elif kind == "ellipse":
        edge_path.add_ellipse(
            (10, 5), (5, 0), ratio=0.7, start_angle=0, end_angle=180, ccw=False
        )
    else:
        pytest.fail(f"unknown kind: {kind}")
    src_path = make_path(hatch)
    assert len(src_path) > 1, "expected non empty path"

    m = Matrix44.scale(sx, sy, 1)
    transformed_hatch = transformed_copy(hatch, m)

    expected_path = src_path.transform(m)
    path_of_transformed_hatch = make_path(transformed_hatch)
    assert (
        have_close_control_vertices(path_of_transformed_hatch, expected_path)
        is True
    )


@pytest.mark.parametrize(
    "sx, sy",
    [(-1, 1), (1, -1), (-1, -1)],
    ids=["mirror-x", "mirror-y", "mirror-xy"],
)
def test_wcs_mirror_transformations_for_all_edge_types(
    sx, sy, all_edge_types_hatch
):
    hatch = all_edge_types_hatch
    src_path = make_path(hatch)
    assert len(src_path) > 1, "expected non empty path"

    m = Matrix44.scale(sx, sy, 1)
    transformed_hatch = transformed_copy(hatch, m)

    expected_path = src_path.transform(m)
    path_of_transformed_hatch = make_path(transformed_hatch)
    assert (
        have_close_control_vertices(path_of_transformed_hatch, expected_path)
        is True
    )



# Copyright (c) 2018-2024 Manfred Moitzi
# License: MIT License
from __future__ import annotations
from typing import TYPE_CHECKING, Iterable, Sequence, Iterator, Optional
import math
import numpy as np

from ezdxf.math import Vec2, UVec
from .bbox import BoundingBox2d
from .construct2d import enclosing_angles
from .circle import ConstructionCircle
from .line import ConstructionRay, ConstructionLine
from .ucs import UCS

if TYPE_CHECKING:
    from ezdxf.entities import Arc
    from ezdxf.layouts import BaseLayout

__all__ = ["ConstructionArc", "arc_chord_length", "arc_segment_count"]

QUARTER_ANGLES = [0, math.pi * 0.5, math.pi, math.pi * 1.5]


class ConstructionArc:
    """Construction tool for 2D arcs.

    :class:`ConstructionArc` represents a 2D arc in the xy-plane, use an
    :class:`UCS` to place a DXF :class:`~ezdxf.entities.Arc` entity in 3D space,
    see method :meth:`add_to_layout`.

    Implements the 2D transformation tools: :meth:`translate`,
    :meth:`scale_uniform` and :meth:`rotate_z`

    Args:
        center: center point as :class:`Vec2` compatible object
        radius: radius
        start_angle: start angle in degrees
        end_angle: end angle in degrees
        is_counter_clockwise: swaps start- and end angle if ``False``

    """

    def __init__(
        self,
        center: UVec = (0, 0),
        radius: float = 1.0,
        start_angle: float = 0.0,
        end_angle: float = 360.0,
        is_counter_clockwise: Optional[bool] = True,
    ):

        self.center = Vec2(center)
        self.radius = radius
        if is_counter_clockwise:
            self.start_angle = start_angle
            self.end_angle = end_angle
        else:
            self.start_angle = end_angle
            self.end_angle = start_angle

    @property
    def start_point(self) -> Vec2:
        """start point of arc as :class:`Vec2`."""
        return self.center + Vec2.from_deg_angle(self.start_angle, self.radius)

    @property
    def end_point(self) -> Vec2:
        """end point of arc as :class:`Vec2`."""
        return self.center + Vec2.from_deg_angle(self.end_angle, self.radius)

    @property
    def is_passing_zero(self) -> bool:
        """Returns ``True`` if the arc passes the x-axis (== 0 degree)."""
        return self.start_angle > self.end_angle

    @property
    def circle(self) -> ConstructionCircle:
        return ConstructionCircle(self.center, self.radius)

    @property
    def bounding_box(self) -> BoundingBox2d:
        """bounding box of arc as :class:`BoundingBox2d`."""
        bbox = BoundingBox2d((self.start_point, self.end_point))
        bbox.extend(self.main_axis_points())
        return bbox

    def angles(self, num: int) -> Iterable[float]:
        """Returns `num` angles from start- to end angle in degrees in
        counter-clockwise order.

        All angles are normalized in the range from [0, 360).

        """
        if num < 2:
            raise ValueError("num >= 2")
        start: float = self.start_angle % 360
        stop: float = self.end_angle % 360
        if stop <= start:
            stop += 360
        for angle in np.linspace(start, stop, num=num, endpoint=True):
            yield angle % 360

    @property
    def angle_span(self) -> float:
        """Returns angle span of arc from start- to end param."""
        end: float = self.end_angle
        if end < self.start_angle:
            end += 360.0
        return end - self.start_angle

    def vertices(self, a: Iterable[float]) -> Iterable[Vec2]:
        """Yields vertices on arc for angles in iterable `a` in WCS as location
        vectors.

        Args:
            a: angles in the range from 0 to 360 in degrees, arc goes
                counter clockwise around the z-axis, WCS x-axis = 0 deg.

        """
        center = self.center
        radius = self.radius

        for angle in a:
            yield center + Vec2.from_deg_angle(angle, radius)

    def flattening(self, sagitta: float) -> Iterable[Vec2]:
        """Approximate the arc by vertices in WCS, argument `sagitta` is the
        max. distance from the center of an arc segment to the center of its
        chord.

        """
        radius: float = abs(self.radius)
        if radius > 0:
            start: float = self.start_angle
            stop: float = self.end_angle
            if math.isclose(start, stop):
                return
            start %= 360
            stop %= 360
            if stop <= start:
                stop += 360
            angle_span: float = math.radians(stop - start)
            count = arc_segment_count(radius, angle_span, sagitta)
            yield from self.vertices(np.linspace(start, stop, count + 1))

    def tangents(self, a: Iterable[float]) -> Iterable[Vec2]:
        """Yields tangents on arc for angles in iterable `a` in WCS as
        direction vectors.

        Args:
            a: angles in the range from 0 to 360 in degrees, arc goes
                counter-clockwise around the z-axis, WCS x-axis = 0 deg.

        """
        for angle in a:
            r: float = math.radians(angle)
            yield Vec2((-math.sin(r), math.cos(r)))

    def main_axis_points(self) -> Iterator[Vec2]:
        center: Vec2 = self.center
        radius: float = self.radius
        start: float = math.radians(self.start_angle)
        end: float = math.radians(self.end_angle)
        for angle in QUARTER_ANGLES:
            if enclosing_angles(angle, start, end):
                yield center + Vec2.from_angle(angle, radius)

    def translate(self, dx: float, dy: float) -> ConstructionArc:
        """Move arc about `dx` in x-axis and about `dy` in y-axis, returns
        `self` (floating interface).

        Args:
            dx: translation in x-axis
            dy: translation in y-axis

        """
        self.center += Vec2(dx, dy)
        return self

    def scale_uniform(self, s: float) -> ConstructionArc:
        """Scale arc inplace uniform about `s` in x- and y-axis, returns
        `self` (floating interface).
        """
        self.radius *= float(s)
        return self

    def rotate_z(self, angle: float) -> ConstructionArc:
        """Rotate arc inplace about z-axis, returns `self`
        (floating interface).

        Args:
            angle: rotation angle in degrees

        """
        self.start_angle += angle
        self.end_angle += angle
        return self

    @property
    def start_angle_rad(self) -> float:
        """Returns the start angle in radians."""
        return math.radians(self.start_angle)

    @property
    def end_angle_rad(self) -> float:
        """Returns the end angle in radians."""
        return math.radians(self.end_angle)

    @staticmethod
    def validate_start_and_end_point(
        start_point: UVec, end_point: UVec
    ) -> tuple[Vec2, Vec2]:
        start_point = Vec2(start_point)
        end_point = Vec2(end_point)
        if start_point == end_point:
            raise ValueError("Start- and end point have to be different points.")
        return start_point, end_point

    @classmethod
    def from_2p_angle(
        cls,
        start_point: UVec,
        end_point: UVec,
        angle: float,
        ccw: bool = True,
    ) -> ConstructionArc:
        """Create arc from two points and enclosing angle. Additional
        precondition: arc goes by default in counter-clockwise orientation from
        `start_point` to `end_point`, can be changed by `ccw` = ``False``.

        Args:
            start_point: start point as :class:`Vec2` compatible object
            end_point: end point as :class:`Vec2` compatible object
            angle: enclosing angle in degrees
            ccw: counter-clockwise direction if ``True``

        """
        _start_point, _end_point = cls.validate_start_and_end_point(
            start_point, end_point
        )
        angle = math.radians(angle)
        if angle == 0:
            raise ValueError("Angle can not be 0.")
        if ccw is False:
            _start_point, _end_point = _end_point, _start_point
        alpha2: float = angle / 2.0
        distance: float = _end_point.distance(_start_point)
        distance2: float = distance / 2.0
        radius: float = distance2 / math.sin(alpha2)
        height: float = distance2 / math.tan(alpha2)
        mid_point: Vec2 = _end_point.lerp(_start_point, factor=0.5)

        distance_vector: Vec2 = _end_point - _start_point
        height_vector: Vec2 = distance_vector.orthogonal().normalize(height)
        center: Vec2 = mid_point + height_vector

        return ConstructionArc(
            center=center,
            radius=radius,
            start_angle=(_start_point - center).angle_deg,
            end_angle=(_end_point - center).angle_deg,
            is_counter_clockwise=True,
        )

    @classmethod
    def from_2p_radius(
        cls,
        start_point: UVec,
        end_point: UVec,
        radius: float,
        ccw: bool = True,
        center_is_left: bool = True,
    ) -> ConstructionArc:
        """Create arc from two points and arc radius.
        Additional precondition: arc goes by default in counter-clockwise
        orientation from `start_point` to `end_point` can be changed
        by `ccw` = ``False``.

        The parameter `center_is_left` defines if the center of the arc is
        left or right of the line from `start_point` to `end_point`.
        Parameter `ccw` = ``False`` swaps start- and end point, which also
        inverts the meaning of ``center_is_left``.

        Args:
            start_point: start point as :class:`Vec2` compatible object
            end_point: end point as :class:`Vec2` compatible object
            radius: arc radius
            ccw: counter-clockwise direction if ``True``
            center_is_left: center point of arc is left of line from start- to
                end point if ``True``

        """
        _start_point, _end_point = cls.validate_start_and_end_point(
            start_point, end_point
        )
        radius = float(radius)
        if radius <= 0:
            raise ValueError("Radius has to be > 0.")
        if ccw is False:
            _start_point, _end_point = _end_point, _start_point

        mid_point: Vec2 = _end_point.lerp(_start_point, factor=0.5)
        distance: float = _end_point.distance(_start_point)
        distance2: float = distance / 2.0
        height: float = math.sqrt(radius**2 - distance2**2)
        center: Vec2 = mid_point + (_end_point - _start_point).orthogonal(
            ccw=center_is_left
        ).normalize(height)

        return ConstructionArc(
            center=center,
            radius=radius,
            start_angle=(_start_point - center).angle_deg,
            end_angle=(_end_point - center).angle_deg,
            is_counter_clockwise=True,
        )

    @classmethod
    def from_3p(
        cls,
        start_point: UVec,
        end_point: UVec,
        def_point: UVec,
        ccw: bool = True,
    ) -> ConstructionArc:
        """Create arc from three points.
        Additional precondition: arc goes in counter-clockwise
        orientation from `start_point` to `end_point`.

        Args:
            start_point: start point as :class:`Vec2` compatible object
            end_point: end point as :class:`Vec2` compatible object
            def_point: additional definition point as :class:`Vec2` compatible
                object
            ccw: counter-clockwise direction if ``True``

        """
        start_point, end_point = cls.validate_start_and_end_point(
            start_point, end_point
        )
        def_point = Vec2(def_point)
        if def_point == start_point or def_point == end_point:
            raise ValueError("def_point has to be different to start- and end point")

        circle = ConstructionCircle.from_3p(start_point, end_point, def_point)
        center = Vec2(circle.center)
        return ConstructionArc(
            center=center,
            radius=circle.radius,
            start_angle=(start_point - center).angle_deg,
            end_angle=(end_point - center).angle_deg,
            is_counter_clockwise=ccw,
        )

    def add_to_layout(
        self, layout: BaseLayout, ucs: Optional[UCS] = None, dxfattribs=None
    ) -> Arc:
        """Add arc as DXF :class:`~ezdxf.entities.Arc` entity to a layout.

        Supports 3D arcs by using an :ref:`UCS`. An :class:`ConstructionArc` is
        always defined in the xy-plane, but by using an arbitrary UCS, the arc
        can be placed in 3D space, automatically OCS transformation included.

        Args:
            layout: destination layout as :class:`~ezdxf.layouts.BaseLayout`
                object
            ucs: place arc in 3D space by :class:`~ezdxf.math.UCS` object
            dxfattribs: additional DXF attributes for the ARC entity

        """
        arc = layout.add_arc(
            center=self.center,
            radius=self.radius,
            start_angle=self.start_angle,
            end_angle=self.end_angle,
            dxfattribs=dxfattribs,
        )
        return arc if ucs is None else arc.transform(ucs.matrix)

    def intersect_ray(
        self, ray: ConstructionRay, abs_tol: float = 1e-10
    ) -> Sequence[Vec2]:
        """Returns intersection points of arc and `ray` as sequence of
        :class:`Vec2` objects.

        Args:
            ray: intersection ray
            abs_tol: absolute tolerance for tests (e.g. test for tangents)

        Returns:
            tuple of :class:`Vec2` objects

            =========== ==================================
            tuple size  Description
            =========== ==================================
            0           no intersection
            1           line intersects or touches the arc at one point
            2           line intersects the arc at two points
            =========== ==================================

        """
        assert isinstance(ray, ConstructionRay)
        return [
            point
            for point in self.circle.intersect_ray(ray, abs_tol)
            if self._is_point_in_arc_range(point)
        ]

    def intersect_line(
        self, line: ConstructionLine, abs_tol: float = 1e-10
    ) -> Sequence[Vec2]:
        """Returns intersection points of arc and `line` as sequence of
        :class:`Vec2` objects.

        Args:
            line: intersection line
            abs_tol: absolute tolerance for tests (e.g. test for tangents)

        Returns:
            tuple of :class:`Vec2` objects

            =========== ==================================
            tuple size  Description
            =========== ==================================
            0           no intersection
            1           line intersects or touches the arc at one point
            2           line intersects the arc at two points
            =========== ==================================

        """
        assert isinstance(line, ConstructionLine)
        return [
            point
            for point in self.circle.intersect_line(line, abs_tol)
            if self._is_point_in_arc_range(point)
        ]

    def intersect_circle(
        self, circle: ConstructionCircle, abs_tol: float = 1e-10
    ) -> Sequence[Vec2]:
        """Returns intersection points of arc and `circle` as sequence of
        :class:`Vec2` objects.

        Args:
            circle: intersection circle
            abs_tol: absolute tolerance for tests

        Returns:
            tuple of :class:`Vec2` objects

            =========== ==================================
            tuple size  Description
            =========== ==================================
            0           no intersection
            1           circle intersects or touches the arc at one point
            2           circle intersects the arc at two points
            =========== ==================================

        """
        assert isinstance(circle, ConstructionCircle)
        return [
            point
            for point in self.circle.intersect_circle(circle, abs_tol)
            if self._is_point_in_arc_range(point)
        ]

    def intersect_arc(
        self, other: ConstructionArc, abs_tol: float = 1e-10
    ) -> Sequence[Vec2]:
        """Returns intersection points of two arcs as sequence of
        :class:`Vec2` objects.

        Args:
            other: other intersection arc
            abs_tol: absolute tolerance for tests

        Returns:
            tuple of :class:`Vec2` objects

            =========== ==================================
            tuple size  Description
            =========== ==================================
            0           no intersection
            1           other arc intersects or touches the arc at one point
            2           other arc intersects the arc at two points
            =========== ==================================

        """
        assert isinstance(other, ConstructionArc)
        return [
            point
            for point in self.circle.intersect_circle(other.circle, abs_tol)
            if self._is_point_in_arc_range(point)
            and other._is_point_in_arc_range(point)
        ]

    def _is_point_in_arc_range(self, point: Vec2) -> bool:
        # The point has to be on the circle defined by the arc, this is not
        # tested here! Helper tools to check intersections.
        start: float = self.start_angle % 360.0
        end: float = self.end_angle % 360.0
        angle: float = (point - self.center).angle_deg % 360.0
        if start > end:  # arc passes 0 degree
            return angle >= start or angle <= end
        return start <= angle <= end


def arc_chord_length(radius: float, sagitta: float) -> float:
    """Returns the chord length for an arc defined by `radius` and
    the `sagitta`_.

    Args:
        radius: arc radius
        sagitta: distance from the center of the arc to the center of its base

    """
    try:
        return 2.0 * math.sqrt(2.0 * radius * sagitta - sagitta * sagitta)
    except ValueError:
        return 0.0


def arc_segment_count(radius: float, angle: float, sagitta: float) -> int:
    """Returns the count of required segments for the approximation
    of an arc for a given maximum `sagitta`_.

    Args:
        radius: arc radius
        angle: angle span of the arc in radians
        sagitta: max. distance from the center of an arc segment to the
            center of its chord

    """
    try:
        chord_length: float = arc_chord_length(radius, sagitta)
        alpha: float = math.asin(chord_length / 2.0 / radius) * 2.0
        return math.ceil(angle / alpha)
    except (ValueError, ZeroDivisionError):
        return 1



#  Copyright (c) 2022, Manfred Moitzi
#  License: MIT License
from __future__ import annotations
from typing import Union, Sequence
import os
import struct
import uuid
import datetime
import enum
import zipfile

from ezdxf.math import Vec3, normal_vector_3p, BoundingBox
from ezdxf.render import MeshTransformer, MeshVertexMerger, MeshBuilder
from ezdxf import __version__


class UnsupportedFileFormat(Exception):
    pass


class ParsingError(Exception):
    pass


def stl_readfile(filename: Union[str, os.PathLike]) -> MeshTransformer:
    """Read ascii or binary `STL`_ file content as :class:`ezdxf.render.MeshTransformer`
    instance.

    Raises:
        ParsingError: vertex parsing error or invalid/corrupt data

    """
    with open(filename, "rb") as fp:
        buffer = fp.read()
    if buffer.startswith(b"solid"):
        s = buffer.decode("ascii", errors="ignore")
        return stl_loads(s)
    else:
        return stl_loadb(buffer)


def stl_loads(content: str) -> MeshTransformer:
    """Load a mesh from an ascii `STL`_ content string as :class:`ezdxf.render.MeshTransformer`
    instance.

    Raises:
        ParsingError: vertex parsing error

    """
    # http://www.fabbers.com/tech/STL_Format#Sct_ASCII
    # This implementation is not very picky and grabs only lines which start
    # with "vertex" or "endloop" and ignores the rest.
    def parse_vertex(line: str) -> Vec3:
        data = line.split()
        return Vec3(float(data[1]), float(data[2]), float(data[3]))

    mesh = MeshVertexMerger()
    face: list[Vec3] = []
    for num, line in enumerate(content.split("\n"), start=1):
        line = line.strip(" \r")
        if line.startswith("vertex"):
            try:
                face.append(parse_vertex(line))
            except (IndexError, ValueError):
                raise ParsingError(f"STL parsing error in line {num}: {line}")
        elif line.startswith("endloop"):
            if len(face) == 3:
                mesh.add_face(face)
            face.clear()
    return MeshTransformer.from_builder(mesh)


def stl_loadb(buffer: bytes) -> MeshTransformer:
    """Load a mesh from a binary `STL`_ data :class:`ezdxf.render.MeshTransformer`
    instance.

    Raises:
        ParsingError: invalid/corrupt data or not a binary STL file

    """
    # http://www.fabbers.com/tech/STL_Format#Sct_ASCII
    index = 80
    n_faces = struct.unpack_from("<I", buffer, index)[0]
    index += 4

    mesh = MeshVertexMerger()
    for _ in range(n_faces):
        try:
            face = struct.unpack_from("<12fH", buffer, index)
        except struct.error:
            raise ParsingError("binary STL parsing error")
        index += 50
        v1 = Vec3(face[3:6])
        v2 = Vec3(face[6:9])
        v3 = Vec3(face[9:12])
        mesh.add_face((v1, v2, v3))
    return MeshTransformer.from_builder(mesh)


def off_readfile(filename: Union[str, os.PathLike]) -> MeshTransformer:
    """Read `OFF`_ file content as :class:`ezdxf.render.MeshTransformer`
    instance.

    Raises:
        ParsingError: vertex or face parsing error

    """
    with open(filename, "rt", encoding="ascii", errors="ignore") as fp:
        content = fp.read()
    return off_loads(content)


def off_loads(content: str) -> MeshTransformer:
    """Load a mesh from a `OFF`_ content string as :class:`ezdxf.render.MeshTransformer`
    instance.

    Raises:
        ParsingError: vertex or face parsing error

    """
    # https://en.wikipedia.org/wiki/OFF_(file_format)
    mesh = MeshVertexMerger()
    lines: list[str] = []
    for line in content.split("\n"):
        line = line.strip(" \n\r")
        # "OFF" in a single line
        if line.startswith("#") or line == "OFF" or line == "":
            continue
        lines.append(line)

    if len(lines) == 0:
        raise ParsingError(f"OFF format parsing error: no data")

    if lines[0].startswith("OFF"):
        # OFF v f e
        lines[0] = lines[0][4:]

    n = lines[0].split()
    try:
        n_vertices, n_faces = int(n[0]), int(n[1])
    except ValueError:
        raise ParsingError(f"OFF format parsing error: {lines[0]}")

    if len(lines) < n_vertices + n_faces:
        raise ParsingError(f"OFF format parsing error: invalid data count")

    for vertex in lines[1 : n_vertices + 1]:
        v = vertex.split()
        try:
            vtx = Vec3(float(v[0]), float(v[1]), float(v[2]))
        except (ValueError, IndexError):
            raise ParsingError(f"OFF format vertex parsing error: {vertex}")
        mesh.vertices.append(vtx)

    index = n_vertices + 1
    face_indices = []
    for face in lines[index : index + n_faces]:
        f = face.split()
        try:
            vertex_count = int(f[0])
        except ValueError:
            raise ParsingError(f"OFF format face parsing error: {face}")
        for index in range(vertex_count):
            try:
                face_indices.append(int(f[1 + index]))
            except (ValueError, IndexError):
                raise ParsingError(
                    f"OFF format face index parsing error: {face}"
                )
        mesh.faces.append(tuple(face_indices))
        face_indices.clear()
    return MeshTransformer.from_builder(mesh)


def obj_readfile(filename: Union[str, os.PathLike]) -> list[MeshTransformer]:
    """Read `OBJ`_ file content as list of :class:`ezdxf.render.MeshTransformer`
    instances.

    Raises:
        ParsingError: vertex or face parsing error

    """
    with open(filename, "rt", encoding="ascii", errors="ignore") as fp:
        content = fp.read()
    return obj_loads(content)


def obj_loads(content: str) -> list[MeshTransformer]:
    """Load one or more meshes from an `OBJ`_ content string as list of
    :class:`ezdxf.render.MeshTransformer` instances.

    Raises:
        ParsingError: vertex parsing error

    """
    # https://en.wikipedia.org/wiki/Wavefront_.obj_file
    # This implementation is not very picky and grabs only lines which start
    # with "v", "g" or "f" and ignores the rest.
    def parse_vertex(l: str) -> Vec3:
        v = l.split()
        return Vec3(float(v[0]), float(v[1]), float(v[2]))

    def parse_face(l: str) -> Sequence[int]:
        return tuple(int(s.split("/")[0]) for s in l.split())

    vertices: list[Vec3] = [Vec3()]  # 1-indexed
    meshes: list[MeshTransformer] = []
    mesh = MeshVertexMerger()
    for num, line in enumerate(content.split("\n"), start=1):
        line = line.strip(" \r")
        if line.startswith("v"):
            try:
                vtx = parse_vertex(line[2:])
            except (IndexError, ValueError):
                raise ParsingError(
                    f"OBJ vertex parsing error in line {num}: {line}"
                )
            vertices.append(vtx)
        elif line.startswith("f"):
            try:
                mesh.add_face(vertices[i] for i in parse_face(line[2:]))
            except ValueError:
                raise ParsingError(
                    f"OBJ face parsing error in line {num}: {line}"
                )
            except IndexError:
                raise ParsingError(
                    f"OBJ face index error (n={len(vertices)}) in line {num}: {line}"
                )

        elif line.startswith("g") and len(mesh.vertices) > 0:
            meshes.append(MeshTransformer.from_builder(mesh))
            mesh = MeshVertexMerger()

    if len(mesh.vertices) > 0:
        meshes.append(MeshTransformer.from_builder(mesh))
    return meshes


def stl_dumps(mesh: MeshBuilder) -> str:
    """Returns the `STL`_ data as string for the given `mesh`.
    This function triangulates the meshes automatically because the `STL`_
    format supports only triangles as faces.

    This function does not check if the mesh obey the
    `STL`_ format `rules <http://www.fabbers.com/tech/STL_Format>`_:

        - The direction of the face normal is outward.
        - The face vertices are listed in counter-clockwise order when looking
          at the object from the outside (right-hand rule).
        - Each triangle must share two vertices with each of its adjacent triangles.
        - The object represented must be located in the all-positive octant
          (non-negative and nonzero).

    """
    lines: list[str] = [f"solid STL generated by ezdxf {__version__}"]
    for face in mesh.tessellation(max_vertex_count=3):
        if len(face) < 3:
            continue
        try:
            n = normal_vector_3p(face[0], face[1], face[2])
        except ZeroDivisionError:
            continue
        n = n.round(8)
        lines.append(f"  facet normal {n.x} {n.y} {n.z}")
        lines.append("    outer loop")
        for v in face:
            lines.append(f"      vertex {v.x} {v.y} {v.z}")
        lines.append("    endloop")
        lines.append("  endfacet")
    lines.append("endsolid\n")
    return "\n".join(lines)


STL_SIGNATURE = (b"STL generated ezdxf" + b" " * 80)[:80]


def stl_dumpb(mesh: MeshBuilder) -> bytes:
    """Returns the `STL`_ binary data as bytes for the given `mesh`.

    For more information see function: :func:`stl_dumps`
    """
    data: list[bytes] = [STL_SIGNATURE, b"0000"]
    count = 0
    for face in mesh.tessellation(max_vertex_count=3):
        try:
            n = normal_vector_3p(face[0], face[1], face[2])
        except ZeroDivisionError:
            continue
        count += 1
        values = list(n.xyz)
        for v in face:
            values.extend(v.xyz)
        values.append(0)
        data.append(struct.pack("<12fH", *values))
    data[1] = struct.pack("<I", count)
    return b"".join(data)


def off_dumps(mesh: MeshBuilder) -> str:
    """Returns the `OFF`_ data as string for the given `mesh`.
    The `OFF`_ format supports ngons as faces.

    """
    lines: list[str] = ["OFF", f"{len(mesh.vertices)} {len(mesh.faces)} 0"]
    for v in mesh.vertices:
        v = v.round(6)
        lines.append(f"{v.x} {v.y} {v.z}")
    for face in mesh.open_faces():
        lines.append(f"{len(face)} {' '.join(str(i) for i in face)}")
    lines[-1] += "\n"
    return "\n".join(lines)


def obj_dumps(mesh: MeshBuilder) -> str:
    """Returns the `OBJ`_ data as string for the given `mesh`.
    The `OBJ`_ format supports ngons as faces.

    """
    lines: list[str] = [f"# OBJ generated by ezdxf {__version__}"]
    for v in mesh.vertices:
        v = v.round(6)
        lines.append(f"v {v.x} {v.y} {v.z}")
    for face in mesh.open_faces():
        # OBJ is 1-index
        lines.append("f " + " ".join(str(i + 1) for i in face))
    lines[-1] += "\n"
    return "\n".join(lines)


def scad_dumps(mesh: MeshBuilder) -> str:
    """Returns the `OpenSCAD`_ `polyhedron`_ definition as string for the given
    `mesh`. `OpenSCAD`_ supports ngons as faces.

    .. Important::

        `OpenSCAD`_ requires the face normals pointing inwards, the method
        :meth:`~ezdxf.render.MeshBuilder.flip_normals` of the
        :class:`~ezdxf.render.MeshBuilder` class can flip the normals
        inplace.

    """
    # polyhedron( points = [ [X0, Y0, Z0], [X1, Y1, Z1], ... ], faces = [ [P0, P1, P2, P3, ...], ... ], convexity = N);   // 2014.03 & later
    lines: list[str] = ["polyhedron(points = ["]
    for v in mesh.vertices:
        v = v.round(6)
        lines.append(f"  [{v.x}, {v.y}, {v.z}],")
    # OpenSCAD accept the last ","
    lines.append("], faces = [")
    for face in mesh.open_faces():
        lines.append("  [" + ", ".join(str(i) for i in face) + "],")
    # OpenSCAD accept the last ","
    lines.append("], convexity = 10);\n")
    return "\n".join(lines)


def ply_dumpb(mesh: MeshBuilder) -> bytes:
    """Returns the `PLY`_ binary data as bytes for the given `mesh`.
    The `PLY`_ format supports ngons as faces.

    """
    if any(len(f) > 255 for f in mesh.faces):
        face_hdr_fmt = b"property list int int vertex_index"
        face_fmt = "<i{}i"
    else:
        face_hdr_fmt = b"property list uchar int vertex_index"
        face_fmt = "<B{}i"

    header: bytes = b"\n".join(
        [
            b"ply",
            b"format binary_little_endian 1.0",
            b"comment generated by ezdxf " + __version__.encode(),
            b"element vertex " + str(len(mesh.vertices)).encode(),
            b"property float x",
            b"property float y",
            b"property float z",
            b"element face " + str(len(mesh.faces)).encode(),
            face_hdr_fmt,
            b"end_header\n",
        ]
    )
    data: list[bytes] = [header]
    for vertex in mesh.vertices:
        data.append(struct.pack("<3f", vertex.x, vertex.y, vertex.z))
    for face in mesh.open_faces():
        count = len(face)
        fmt = face_fmt.format(count)
        data.append(struct.pack(fmt, count, *face))
    return b"".join(data)


def ifc_guid() -> str:
    return _guid_compress(uuid.uuid4().hex)


def _guid_compress(g: str) -> str:
    # https://github.com/IfcOpenShell/IfcOpenShell/blob/master/src/ifcopenshell-python/ifcopenshell/guid.py#L56
    chars = "0123456789ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz_$"
    bs = [int(g[i : i + 2], 16) for i in range(0, len(g), 2)]

    def b64(v: int, size: int = 4):
        return "".join(
            [chars[(v // (64**i)) % 64] for i in range(size)][::-1]
        )

    return "".join(
        [b64(bs[0], 2)]
        + [
            b64((bs[i] << 16) + (bs[i + 1] << 8) + bs[i + 2])
            for i in range(1, 16, 3)
        ]
    )


class IfcEntityType(enum.Enum):
    POLYGON_FACE_SET = "POLY_FACE_SET"
    CLOSED_SHELL = "CLOSED_SHELL"
    OPEN_SHELL = "OPEN_SHELL"


class Records:
    def __init__(self) -> None:
        # Record numbers are 1-based, record index in list is 0-based!
        # records[0] is record #1
        self.records: list[str] = []

    @property
    def last_num(self) -> int:  # list index
        return len(self.records)

    @property
    def prev_num(self) -> int:  # list index
        return self.last_num - 1

    @property
    def next_num(self) -> int:  # list index
        return self.last_num + 1

    def add(self, record: str, num: int = 0) -> str:
        assert record.endswith(");"), "invalid structure"
        self.records.append(record)
        if num != 0 and num != len(self.records):
            raise ValueError("unexpected record number")
        num = len(self.records)
        return f"#{num}"

    def add_dummy(self):
        self.records.append("")

    def get(self, num: int) -> str:
        return self.records[num - 1]

    def update_all(self, tag: str, record_num: str):
        for index, record in enumerate(self.records):
            if tag in record:
                self.update_record(index, tag, record_num)

    def update_record(self, index, tag: str, record_num: str):
        self.records[index] = self.records[index].replace(tag, record_num)

    def dumps(self) -> str:
        return "\n".join(
            f"#{num+1}= {data}" for num, data in enumerate(self.records) if data
        )


def ifc4_dumps(
    mesh: MeshBuilder,
    entity_type=IfcEntityType.POLYGON_FACE_SET,
    *,
    layer: str = "MeshExport",
    color: tuple[float, float, float] = (1.0, 1.0, 1.0),
) -> str:
    """Returns the `IFC4`_ string for the given `mesh`. The caller is
    responsible for checking if the mesh is a closed or open surface
    (e.g. :code:`mesh.diagnose().euler_characteristic == 2`) and using the
    appropriate entity type.

    Args:
        mesh: :class:`~ezdxf.render.MeshBuilder`
        entity_type: :class:`IfcEntityType`
        layer: layer name as string
        color: entity color as RGB tuple, values in the range [0,1]

    .. warning::

        `IFC4`_ is a very complex data format and this is a minimal effort
        exporter, so the exported data may not be importable by all CAD
        applications.

        The exported `IFC4`_ data can be imported by the following applications:

        - BricsCAD
        - FreeCAD (IfcOpenShell)
        - Allplan
        - Tekla BIMsight

    """

    def make_header():
        date = datetime.datetime.now().isoformat()[:-7]
        return f"""ISO-10303-21;
HEADER;
FILE_DESCRIPTION(('ViewDefinition [CoordinationView_V2.0]'),'2;1');
FILE_NAME('undefined.ifc','{date}',('Undefined'),('Undefined'),'ezdxf {__version__}','ezdxf {__version__}','Undefined');
FILE_SCHEMA(('IFC4'));
ENDSEC;

DATA;
"""

    def make_data_records() -> Records:
        records = Records()
        # fmt: off
        if entity_type == IfcEntityType.POLYGON_FACE_SET:
            kind = "SurfaceModel"
        elif entity_type == IfcEntityType.OPEN_SHELL:
            kind = "SurfaceModel"
        elif entity_type == IfcEntityType.CLOSED_SHELL:
            kind = "Brep"
        else:
            raise ValueError(f"invalid entity type: {entity_type}")
        # for simplicity the first part has absolute record numbering:
        records.add(f"IFCPROJECT('{ifc_guid()}',#2,'MeshExport',$,$,$,$,(#7),#13);", 1)
        records.add("IFCOWNERHISTORY(#3,#6,$,$,$,$,$,0);", 2)
        records.add("IFCPERSONANDORGANIZATION(#4,#5,$);", 3)
        records.add("IFCPERSON($,$,'Undefined',$,$,$,$,$);", 4)
        records.add("IFCORGANIZATION($,'Undefined',$,$,$);", 5)
        records.add(f"IFCAPPLICATION(#5,'{__version__}','ezdxf','ezdxf');", 6)
        records.add("IFCGEOMETRICREPRESENTATIONCONTEXT($,'Model',3,1.000000000000000E-05,#8,#12);", 7)
        records.add("IFCAXIS2PLACEMENT3D(#9,#10,#11);", 8)
        records.add("IFCCARTESIANPOINT((0.,0.,0.));", 9)
        records.add("IFCDIRECTION((0.,0.,1.));", 10)
        records.add("IFCDIRECTION((1.,0.,0.));", 11)
        records.add("IFCDIRECTION((1.,0.));", 12)
        records.add("IFCUNITASSIGNMENT((#14,#15,#16,#17,#18,#19));", 13)
        records.add("IFCSIUNIT(*,.LENGTHUNIT.,$,.METRE.);", 14)
        records.add("IFCSIUNIT(*,.AREAUNIT.,$,.SQUARE_METRE.);", 15)
        records.add("IFCSIUNIT(*,.VOLUMEUNIT.,$,.CUBIC_METRE.);", 16)
        records.add("IFCSIUNIT(*,.PLANEANGLEUNIT.,$,.RADIAN.);", 17)
        records.add("IFCSIUNIT(*,.TIMEUNIT.,$,.SECOND.);", 18)
        records.add("IFCSIUNIT(*,.MASSUNIT.,$,.GRAM.);", 19)
        records.add("IFCLOCALPLACEMENT($,#8);", 20)
        building = records.add(f"IFCBUILDING('{ifc_guid()}',#2,'MeshExport',$,$, #20,$,$,.ELEMENT.,$,$,$);", 21)
        records.add_dummy()  # I don't wanna redo the absolute record numbering
        proxy = records.add(f"IFCBUILDINGELEMENTPROXY('{ifc_guid()}',#2,$,$,$,#24,#29,$,$);", 23)
        records.add("IFCLOCALPLACEMENT(#25,#26);", 24)
        records.add("IFCLOCALPLACEMENT(#20,#8);", 25)
        records.add("IFCAXIS2PLACEMENT3D(#27,#10,#28);", 26)
        records.add(f"IFCCARTESIANPOINT(({emin.x},{emin.y},{emin.z}));", 27)
        records.add("IFCDIRECTION((1.,0.,0.));", 28)
        records.add("IFCPRODUCTDEFINITIONSHAPE($,$,(#30));", 29)
        shape = records.add(f"IFCSHAPEREPRESENTATION(#31,'Body','{kind}',($ENTITY$));", 30)
        records.add("IFCGEOMETRICREPRESENTATIONSUBCONTEXT('Body','Model',*,*,*,*,#7,$,.MODEL_VIEW.,$);", 31)

        # from here on only relative record numbering:
        entity = "#0"
        if entity_type == IfcEntityType.POLYGON_FACE_SET:
            entity = make_polygon_face_set(records)
        elif entity_type == IfcEntityType.CLOSED_SHELL:
            entity = make_shell(records)
        elif entity_type == IfcEntityType.OPEN_SHELL:
            entity = make_shell(records)
        color_str = f"IFCCOLOURRGB($,{color[0]:.3f},{color[1]:.3f},{color[2]:.3f});"
        records.add(color_str)
        records.add(f"IFCSURFACESTYLESHADING(#{records.prev_num+1},0.);")
        records.add(f"IFCSURFACESTYLE($,.POSITIVE.,(#{records.prev_num+1}));")
        records.add(f"IFCPRESENTATIONSTYLEASSIGNMENT((#{records.prev_num+1}));")
        records.add(f"IFCSTYLEDITEM($ENTITY$,(#{records.prev_num+1}),$);")
        records.add(f"IFCPRESENTATIONLAYERWITHSTYLE('{layer}',$,({shape}),$,.T.,.F.,.F.,(#{records.next_num+1}));")
        records.add(f"IFCSURFACESTYLE($,.POSITIVE.,(#{records.next_num+1}));")
        records.add(f"IFCSURFACESTYLESHADING(#{records.next_num+1},0.);")
        records.add(color_str)
        records.add(f"IFCRELCONTAINEDINSPATIALSTRUCTURE('{ifc_guid()}',#2,$,$,({proxy}),{building});")
        records.add(f"IFCRELAGGREGATES('{ifc_guid()}',#2,$,$,#1,({building}));")
        # fmt: on
        records.update_all("$ENTITY$", entity)
        return records

    def make_polygon_face_set(records: Records) -> str:
        entity = records.add(
            f"IFCPOLYGONALFACESET(#{records.next_num+1},$,($FACES$), $);"
        )
        entity_num = records.last_num
        vertices = ",".join([str(v.xyz) for v in mesh.vertices])
        records.add(f"IFCCARTESIANPOINTLIST3D(({vertices}));")
        face_records: list[str] = []
        for face in mesh.open_faces():
            indices = ",".join(str(i + 1) for i in face)  # 1-based indexing
            face_records.append(
                records.add(f"IFCINDEXEDPOLYGONALFACE(({indices}));")
            )
        # list index required
        records.update_record(entity_num - 1, "$FACES$", ",".join(face_records))
        return entity

    def make_shell(records: Records) -> str:
        if entity_type == IfcEntityType.CLOSED_SHELL:
            entity = records.add(f"IFCFACETEDBREP(#{records.next_num+1});")
            records.add(f"IFCCLOSEDSHELL(($FACES$));")
        elif entity_type == IfcEntityType.OPEN_SHELL:
            entity = records.add(
                f"IFCSHELLBASEDSURFACEMODEL((#{records.next_num + 1}));"
            )
            records.add(f"IFCOPENSHELL(($FACES$));")
        else:
            raise ValueError(f"invalid entity type: {entity_type}")
        shell_num = records.last_num
        # add vertices
        first_vertex = records.next_num
        for v in mesh.vertices:
            records.add(f"IFCCARTESIANPOINT({str(v.xyz)});")
        # add faces
        face_records: list[str] = []
        for face in mesh.open_faces():
            vertices = ",".join("#" + str(first_vertex + i) for i in face)
            records.add(f"IFCPOLYLOOP(({vertices}));")
            records.add(f"IFCFACEOUTERBOUND(#{records.prev_num+1},.T.);")
            face_records.append(
                records.add(f"IFCFACE((#{records.prev_num+1}));")
            )
        # list index required
        records.update_record(shell_num - 1, "$FACES$", ",".join(face_records))
        return entity

    if len(mesh.vertices) == 0:
        return ""
    bbox = BoundingBox(mesh.vertices)
    emin = Vec3()
    assert bbox.extmin is not None
    if bbox.extmin.x < 0 or bbox.extmin.y < 0 or bbox.extmin.z < 0:
        # Allplan (IFC?) requires all mesh vertices in the all-positive octant
        # (non-negative).  Record #27 does the final placement at the correct
        # location.
        emin = bbox.extmin
        mesh = MeshTransformer.from_builder(mesh)
        mesh.translate(-emin.x, -emin.y, -emin.z)

    header = make_header()
    data = make_data_records()
    return header + data.dumps() + "\nENDSEC;\nEND-ISO-10303-21;\n"


def export_ifcZIP(
    filename: Union[str, os.PathLike],
    mesh: MeshBuilder,
    entity_type=IfcEntityType.POLYGON_FACE_SET,
    *,
    layer: str = "MeshExport",
    color: tuple[float, float, float] = (1.0, 1.0, 1.0),
):
    """Export the given `mesh` as zip-compressed `IFC4`_ file. The filename
    suffix should be ``.ifcZIP``. For more information see function
    :func:`ifc4_dumps`.

    Args:
        filename: zip filename, the data file has the same name with suffix ``.ifc``
        mesh: :class:`~ezdxf.render.MeshBuilder`
        entity_type: :class:`IfcEntityType`
        layer: layer name as string
        color: entity color as RGB tuple, values in the range [0,1]

    Raises:
        IOError: IO error when opening the zip-file for writing

    """
    name = os.path.basename(filename) + ".ifc"
    zf = zipfile.ZipFile(filename, mode="w", compression=zipfile.ZIP_DEFLATED)
    try:
        zf.writestr(
            name, ifc4_dumps(mesh, entity_type, layer=layer, color=color)
        )
    finally:
        zf.close()



# Copyright (c) 2021-2023 Manfred Moitzi
# License: MIT License
import pytest

from ezdxf.entities import is_graphic_entity, is_dxf_object
from ezdxf.lldxf.tagwriter import TagCollector, basic_tags_from_text
from ezdxf.entities.dxfentity import DXFTagStorage
from ezdxf.protocols import SupportsVirtualEntities, query_virtual_entities


@pytest.fixture
def entity():
    return DXFTagStorage.from_text(THE_KNOWN_UNKNOWN)


def test_default_attribs(entity):
    assert entity.dxftype() == "MTEXT"
    assert entity.dxf.handle == "278"
    assert entity.dxf.owner == "1F"
    assert entity.base_class[0] == (0, "MTEXT")
    assert entity.base_class[1] == (5, "278")


def test_wrapped_mtext_is_a_graphic_entity(entity):
    assert entity.is_graphic_entity is True
    assert is_graphic_entity(entity) is True


def test_wrapped_mtext_is_not_a_dxf_object(entity):
    assert is_dxf_object(entity) is False


def test_wrapped_mtext_returns_dxfattribs_dict(entity):
    dxfattribs = entity.graphic_properties()
    assert len(dxfattribs) == 1
    assert dxfattribs["layer"] == "0"


def test_dxf_tag_storage_is_a_non_graphical_entity_by_default():
    assert DXFTagStorage().is_graphic_entity is False
    assert is_graphic_entity(DXFTagStorage()) is False


def test_dxf_tag_storage_returns_empty_dxfattribs_dict_for_non_graphical_objetcs():
    assert len(DXFTagStorage().graphic_properties()) == 0


def test_dxf_export(entity):
    control_tags = basic_tags_from_text(THE_KNOWN_UNKNOWN)
    collector = TagCollector()
    entity.export_dxf(collector)
    result = collector.tags
    assert result == control_tags


def test_virtual_entities(entity):
    assert len(list(entity.virtual_entities())) == 0


def test_supports_virtual_entities_protocol(entity):
    assert isinstance(entity, SupportsVirtualEntities) is True
    assert len(query_virtual_entities(entity)) == 0


THE_KNOWN_UNKNOWN = r"""0
MTEXT
5
278
330
1F
100
AcDbEntity
8
0
100
AcDbMText
10
2762.147
20
2327.073
30
0.0
40
2.5
41
18.851
46
0.0
71
1
72
5
1
{\fArial|b0|i0|c162|p34;CHANGE;\P\P\PTEXT}
73
1
44
1.0
101
Embedded Object
70
1
10
1.0
20
0.0
30
0.0
11
2762.147
21
2327.073
31
0.0
40
18.851
41
0.0
42
15.428
43
15.042
71
2
72
1
44
18.851
45
12.5
73
0
74
0
46
0.0
"""

if __name__ == "__main__":
    pytest.main([__file__])



# Copyright (c) 2021-2022, Manfred Moitzi
# License: MIT License
from __future__ import annotations
from typing import Union, Iterable, TYPE_CHECKING
from pathlib import Path

from ezdxf.lldxf import loader
from ezdxf.lldxf.types import DXFTag
from ezdxf.lldxf.tagger import ascii_tags_loader
from ezdxf.lldxf.validator import is_dxf_file
from ezdxf.filemanagement import dxf_file_info

if TYPE_CHECKING:
    from ezdxf.eztypes import SectionDict


def raw_structure_loader(filename: Union[str, Path]) -> SectionDict:
    """Load content of ASCII DXF file `filename` as SectionDict, all tags are in
    raw format with the group code as integer and the value as string: (int, str).

    """
    tagger = get_tag_loader(filename)
    return loader.load_dxf_structure(tagger)


def get_tag_loader(
    filename: Union[str, Path], errors: str = "ignore"
) -> Iterable[DXFTag]:

    filename = str(filename)
    if not is_dxf_file(filename):
        raise IOError(f"File '{filename}' is not an ASCII DXF file.")

    info = dxf_file_info(filename)
    with open(filename, mode="rt", encoding=info.encoding, errors=errors) as fp:
        return list(ascii_tags_loader(fp, skip_comments=True))



#  Copyright (c) 2023, Manfred Moitzi
#  License: MIT License
from __future__ import annotations
from typing import Iterator, Iterable
import string

from .deps import Vec2, NULLVEC2
from .properties import RGB
from .plotter import Plotter
from .tokenizer import Command, pe_decode


class Interpreter:
    """The :class:`Interpreter` is the frontend for the :class:`Plotter` class.
    The :meth:`run` methods interprets the low level HPGL commands from the
    :func:`hpgl2_commands` parser and sends the commands to the virtual plotter
    device, which sends his output to a low level :class:`Backend` class.

    Most CAD application send a very restricted subset of commands to plotters,
    mostly just polylines and filled polygons. Implementing the whole HPGL/2 command set
    is not worth the effort - unless reality proofs otherwise.

    Not implemented commands:

        - the whole character group - text is send as filled polygons or polylines
        - configuration group: IN, DF, RO, IW - the plotter is initialized by creating a
          new plotter and page rotation is handled by the add-on itself
        - polygon group: EA, ER, EW, FA, RR, WG, the rectangle and wedge commands
        - line and fill attributes group: LA, RF, SM, SV, TR, UL, WU, linetypes and
          hatch patterns are decomposed into simple lines by CAD applications

    Args:
        plotter: virtual :class:`Plotter` device

    """
    def __init__(self, plotter: Plotter) -> None:
        self.errors: list[str] = []
        self.not_implemented_commands: set[str] = set()
        self._disabled_commands: set[str] = set()
        self.plotter = plotter

    def add_error(self, error: str) -> None:
        self.errors.append(error)

    def run(self, commands: list[Command]) -> None:
        """Interprets the low level HPGL commands from the :func:`hpgl2_commands` parser
        and sends the commands to the virtual plotter device.
        """
        for name, args in commands:
            if name in self._disabled_commands:
                continue
            method = getattr(self, f"cmd_{name.lower()}", None)
            if method:
                method(args)
            elif name[0] in string.ascii_letters:
                self.not_implemented_commands.add(name)

    def disable_commands(self, commands: Iterable[str]) -> None:
        """Disable commands manually, like the scaling command ["SC", "IP", "IR"].
        This is a feature for experts, because disabling commands which changes the pen
        location may distort or destroy the plotter output.
        """
        self._disabled_commands.update(commands)

    # Configure pens, line types, fill types
    def cmd_ft(self, args: list[bytes]):
        """Set fill type."""
        fill_type = 1
        spacing = 0.0
        angle = 0.0
        values = tuple(to_floats(args))
        arg_count = len(values)
        if arg_count > 0:
            fill_type = int(values[0])
        if arg_count > 1:
            spacing = values[1]
        if arg_count > 2:
            angle = values[2]
        self.plotter.set_fill_type(fill_type, spacing, angle)

    def cmd_pc(self, args: list[bytes]):
        """Set pen color as RGB tuple."""
        values = list(to_ints(args))
        if len(values) == 4:
            index, r, g, b = values
            self.plotter.set_pen_color(index, RGB(r, g, b))
        else:
            self.add_error("invalid arguments for PC command")

    def cmd_pw(self, args: list[bytes]):
        """Set pen width."""
        arg_count = len(args)
        if arg_count:
            width = to_float(args[0], 0.35)
        else:
            self.add_error("invalid arguments for PW command")
            return
        index = -1
        if arg_count > 1:
            index = to_int(args[1], index)
        self.plotter.set_pen_width(index, width)

    def cmd_sp(self, args: list[bytes]):
        """Select pen."""
        if len(args):
            self.plotter.set_current_pen(to_int(args[0], 1))

    def cmd_np(self, args: list[bytes]):
        """Set number of pens."""
        if len(args):
            self.plotter.set_max_pen_count(to_int(args[0], 2))

    def cmd_ip(self, args: list[bytes]):
        """Set input points p1 and p2 absolute."""
        if len(args) == 0:
            self.plotter.reset_scaling()
            return

        points = to_points(to_floats(args))
        if len(points) > 1:
            self.plotter.set_scaling_points(points[0], points[1])
        else:
            self.add_error("invalid arguments for IP command")

    def cmd_ir(self, args: list[bytes]):
        """Set input points p1 and p2 in percentage of page size."""
        if len(args) == 0:
            self.plotter.reset_scaling()
            return

        values = list(to_floats(args))
        if len(values) == 2:
            xp1 = clamp(values[0], 0.0, 100.0)
            yp1 = clamp(values[1], 0.0, 100.0)
            self.plotter.set_scaling_points_relative_1(xp1 / 100.0, yp1 / 100.0)
        elif len(values) == 4:
            xp1 = clamp(values[0], 0.0, 100.0)
            yp1 = clamp(values[1], 0.0, 100.0)
            xp2 = clamp(values[2], 0.0, 100.0)
            yp2 = clamp(values[3], 0.0, 100.0)
            self.plotter.set_scaling_points_relative_2(
                xp1 / 100.0, yp1 / 100.0, xp2 / 100.0, yp2 / 100.0
            )
        else:
            self.add_error("invalid arguments for IP command")

    def cmd_sc(self, args: list[bytes]):
        if len(args) == 0:
            self.plotter.reset_scaling()
            return
        values = list(to_floats(args))
        if len(values) < 4:
            self.add_error("invalid arguments for SC command")
            return
        scaling_type = 0
        if len(values) > 4:
            scaling_type = int(values[4])
        if scaling_type == 1:  # isotropic
            left = 50.0
            if len(values) > 5:
                left = clamp(values[5], 0.0, 100.0)
            bottom = 50.0
            if len(values) > 6:
                bottom = clamp(values[6], 0.0, 100.0)
            self.plotter.set_isotropic_scaling(
                values[0],
                values[1],
                values[2],
                values[3],
                left,
                bottom,
            )
        elif scaling_type == 2:  # point factor
            self.plotter.set_point_factor(
                Vec2(values[0], values[2]), values[1], values[3]
            )
        else:  # anisotropic
            self.plotter.set_anisotropic_scaling(
                values[0], values[1], values[2], values[3]
            )

    def cmd_mc(self, args: list[bytes]):
        status = 0
        if len(args):
            status = to_int(args[0], status)
            self.plotter.set_merge_control(bool(status))

    def cmd_ps(self, args: list[bytes]):
        length = 1189  # A0
        height = 841
        count = len(args)
        if count:
            length = to_int(args[0], length)
            height = int(length * 1.5)
        if count > 1:
            height = to_int(args[1], height)
        self.plotter.setup_page(length, height)

    # pen movement:
    def cmd_pd(self, args: list[bytes]):
        """Lower pen down and plot lines."""
        self.plotter.pen_down()
        if len(args):
            self.plotter.plot_polyline(to_points(to_floats(args)))

    def cmd_pu(self, args: list[bytes]):
        """Lift pen up and move pen."""
        self.plotter.pen_up()
        if len(args):
            self.plotter.plot_polyline(to_points(to_floats(args)))

    def cmd_pa(self, args: list[bytes]):
        """Place pen absolute. Plots polylines if pen is down."""
        self.plotter.set_absolute_mode()
        if len(args):
            self.plotter.plot_polyline(to_points(to_floats(args)))

    def cmd_pr(self, args: list[bytes]):
        """Place pen relative.Plots polylines if pen is down."""
        self.plotter.set_relative_mode()
        if len(args):
            self.plotter.plot_polyline(to_points(to_floats(args)))

    # plot commands:
    def cmd_ci(self, args: list[bytes]):
        """Plot full circle."""
        arg_count = len(args)
        if not arg_count:
            self.add_error("invalid arguments for CI command")
            return
        self.plotter.push_pen_state()
        # implicit pen down!
        self.plotter.pen_down()
        radius = to_float(args[0], 1.0)
        chord_angle = 5.0
        if arg_count > 1:
            chord_angle = to_float(args[1], chord_angle)
        self.plotter.plot_abs_circle(radius, chord_angle)
        self.plotter.pop_pen_state()

    def cmd_aa(self, args: list[bytes]):
        """Plot arc absolute."""
        if len(args) < 3:
            self.add_error("invalid arguments for AR command")
            return
        self._arc_out(args, self.plotter.plot_abs_arc)

    def cmd_ar(self, args: list[bytes]):
        """Plot arc relative."""
        if len(args) < 3:
            self.add_error("invalid arguments for AR command")
            return
        self._arc_out(args, self.plotter.plot_rel_arc)

    @staticmethod
    def _arc_out(args: list[bytes], output_method):
        """Plot arc"""
        arg_count = len(args)
        if arg_count < 3:
            return
        x = to_float(args[0])
        y = to_float(args[1])
        sweep_angle = to_float(args[2])
        chord_angle = 5.0
        if arg_count > 3:
            chord_angle = to_float(args[3], chord_angle)
        output_method(Vec2(x, y), sweep_angle, chord_angle)

    def cmd_at(self, args: list[bytes]):
        """Plot arc absolute from three points."""
        if len(args) < 4:
            self.add_error("invalid arguments for AT command")
            return
        self._arc_3p_out(args, self.plotter.plot_abs_arc_three_points)

    def cmd_rt(self, args: list[bytes]):
        """Plot arc relative from three points."""
        if len(args) < 4:
            self.add_error("invalid arguments for RT command")
            return
        self._arc_3p_out(args, self.plotter.plot_rel_arc_three_points)

    @staticmethod
    def _arc_3p_out(args: list[bytes], output_method):
        """Plot arc from three points"""
        arg_count = len(args)
        if arg_count < 4:
            return
        points = to_points(to_floats(args))
        if len(points) < 2:
            return
        chord_angle = 5.0
        if arg_count > 4:
            chord_angle = to_float(args[4], chord_angle)
        try:
            output_method(points[0], points[1], chord_angle)
        except ZeroDivisionError:
            pass

    def cmd_bz(self, args: list[bytes]):
        """Plot cubic Bezier curves with absolute user coordinates."""
        self._bezier_out(args, self.plotter.plot_abs_cubic_bezier)

    def cmd_br(self, args: list[bytes]):
        """Plot cubic Bezier curves with relative user coordinates."""
        self._bezier_out(args, self.plotter.plot_rel_cubic_bezier)

    @staticmethod
    def _bezier_out(args: list[bytes], output_method):
        kind = 0
        ctrl1 = NULLVEC2
        ctrl2 = NULLVEC2
        for point in to_points(to_floats(args)):
            if kind == 0:
                ctrl1 = point
            elif kind == 1:
                ctrl2 = point
            elif kind == 2:
                end = point
                output_method(ctrl1, ctrl2, end)
            kind = (kind + 1) % 3

    def cmd_pe(self, args: list[bytes]):
        """Plot Polyline Encoded."""
        if len(args):
            data = args[0]
        else:
            self.add_error("invalid arguments for PE command")
            return

        plotter = self.plotter
        # The last pen up/down state remains after leaving the PE command.
        pen_down = True
        # Ignores and preserves the current absolute/relative mode of the plotter.
        absolute = False
        frac_bin_bits = 0
        base = 64
        index = 0
        length = len(data)
        point_queue: list[Vec2] = []

        while index < length:
            char = data[index]
            if char in b":<>=7":
                index += 1
                if char == 58:  # ":" - select pen
                    values, index = pe_decode(data, base=base, start=index)
                    plotter.set_current_pen(int(values[0]))
                    if len(values) > 1:
                        point_queue.extend(to_points(values[1:]))
                elif char == 60:  # "<" -  pen up and goto coordinates
                    pen_down = False
                elif char == 62:  # ">" - fractional data
                    values, index = pe_decode(data, base=base, start=index)
                    frac_bin_bits = int(values[0])
                    if len(values) > 1:
                        point_queue.extend(to_points(values[1:]))
                elif char == 61:  # "=" - next coordinates are absolute
                    absolute = True
                elif char == 55:  # "7" - 7-bit mode
                    base = 32
            else:
                values, index = pe_decode(
                    data, frac_bits=frac_bin_bits, base=base, start=index
                )
                point_queue.extend(to_points(values))

            if point_queue:
                plotter.pen_down()
                if absolute:
                    # next point is absolute: make relative
                    point_queue[0] = point_queue[0] - plotter.user_location
                if not pen_down:
                    target = point_queue.pop(0)
                    plotter.move_to_rel(target)
                    if not point_queue:  # last point in queue
                        plotter.pen_up()
                if point_queue:
                    plotter.plot_rel_polyline(point_queue)
                    point_queue.clear()
                pen_down = True
                absolute = False

    # polygon mode:
    def cmd_pm(self, args: list[bytes]) -> None:
        """Enter/Exit polygon mode."""
        status = 0
        if len(args):
            status = to_int(args[0], status)
        if status == 2:
            self.plotter.exit_polygon_mode()
        else:
            self.plotter.enter_polygon_mode(status)

    def cmd_fp(self, args: list[bytes]) -> None:
        """Plot filled polygon."""
        fill_method = 0
        if len(args):
            fill_method = one_of(to_int(args[0], fill_method), (0, 1))
        self.plotter.fill_polygon(fill_method)

    def cmd_ep(self, _) -> None:
        """Plot edged polygon."""
        self.plotter.edge_polygon()


def to_floats(args: Iterable[bytes]) -> Iterator[float]:
    for arg in args:
        try:
            yield float(arg)
        except ValueError:
            pass


def to_ints(args: Iterable[bytes]) -> Iterator[int]:
    for arg in args:
        try:
            yield int(arg)
        except ValueError:
            pass


def to_points(values: Iterable[float]) -> list[Vec2]:
    points: list[Vec2] = []
    append_point = False
    buffer: float = 0.0
    for value in values:
        if append_point:
            points.append(Vec2(buffer, value))
            append_point = False
        else:
            buffer = value
            append_point = True
    return points


def to_float(s: bytes, default=0.0) -> float:
    try:
        return float(s)
    except ValueError:
        return default


def to_int(s: bytes, default=0) -> int:
    try:
        return int(s)
    except ValueError:
        return default


def clamp(v, v_min, v_max):
    return max(min(v_max, v), v_min)


def one_of(value, choice):
    if value in choice:
        return value
    return choice[0]



#  Copyright (c) 2021, Manfred Moitzi
#  License: MIT License

import pytest
import ezdxf
from ezdxf.math import Vec2
from ezdxf.render import mleader
from ezdxf.entities import MText, MultiLeader, Insert


@pytest.fixture(scope="module")
def doc():
    return ezdxf.new()


def make_multi_leader(doc) -> MultiLeader:
    style = doc.mleader_styles.get("Standard")
    ml = MultiLeader.new(doc=doc)
    ml.dxf.style_handle = style.dxf.handle
    return ml


class TestRenderEngine:
    """The RenderEngine renders DXF primitives from a MULTILEADER entity."""

    @pytest.fixture
    def ml_mtext(self, doc):
        ml = make_multi_leader(doc)
        builder = mleader.MultiLeaderMTextBuilder(ml)
        builder.set_content("line")
        builder.build(insert=Vec2(0, 0))
        return ml

    def test_add_mtext_content(self, ml_mtext):
        engine = mleader.RenderEngine(ml_mtext, ml_mtext.doc)
        engine.add_mtext_content()
        assert isinstance(engine.entities[0], MText)


if __name__ == "__main__":
    pytest.main([__file__])



# Copyright (c) 2018-2019, Manfred Moitzi
# License: MIT License
import pytest
import os
import ezdxf
from ezdxf.lldxf.const import versions_supported_by_new


@pytest.fixture(params=versions_supported_by_new)
def drawing(request):
    return ezdxf.new(request.param)


def add_table_entries(dwg):
    dwg.layers.new("MOZMAN-LAYER")
    dwg.styles.new("MOZMAN-STY")
    dwg.linetypes.new("MOZMAN-LTY", {"pattern": [1.0, 0.5, -0.5]})
    dwg.dimstyles.new("MOZMAN-DIMSTY")
    dwg.views.new("MOZMAN-VIEW")
    dwg.viewports.new("MOZMAN-VPORT")
    dwg.ucs.new("MOZMAN-UCS")
    dwg.appids.new("MOZMANAPP")


def test_adding_table_entries(drawing, tmpdir):
    add_table_entries(drawing)
    filename = str(tmpdir.join("table_entries_%s.dxf" % drawing.dxfversion))
    try:
        drawing.saveas(filename)
    except ezdxf.DXFError as e:
        pytest.fail(
            "DXFError: {0} for DXF version {1}".format(
                str(e), drawing.dxfversion
            )
        )
    assert os.path.exists(filename)



#  Copyright (c) 2021, Manfred Moitzi
#  License: MIT License

import pytest
import ezdxf
from ezdxf.lldxf import const
from ezdxf.math import (
    required_fit_points,
    required_control_points,
    required_knot_values,
    uniform_knot_vector,
)
from ezdxf.audit import Auditor, AuditError


@pytest.fixture(scope="module")
def doc():
    return ezdxf.new()


@pytest.fixture
def auditor(doc):
    return Auditor(doc)


@pytest.fixture
def spline(doc):
    msp = doc.modelspace()
    return msp.add_spline()


@pytest.mark.parametrize(
    "order, expected", [(0, 2), (1, 2), (2, 2), (3, 3), (4, 4), (5, 5)]
)
def test_required_fit_points_without_given_end_tangents(order, expected):
    assert required_fit_points(order, tangents=False) == expected


@pytest.mark.parametrize(
    "order, expected", [(0, 2), (1, 2), (2, 2), (3, 2), (4, 2), (5, 3)]
)
def test_required_fit_points_with_given_end_tangents(order, expected):
    assert required_fit_points(order, tangents=True) == expected


@pytest.mark.parametrize(
    "order, expected",
    [
        (0, 2),
        (1, 2),
        (2, 2),
        (3, 3),
        (4, 4),
    ],
)
def test_required_control_points_calculation(order, expected):
    assert required_control_points(order) == expected


def test_any_points_present(auditor, spline):
    spline.audit(auditor)
    assert len(auditor.fixes) == 1
    assert auditor.fixes[0].code == AuditError.INVALID_SPLINE_DEFINITION

    # Test if invalid spline will be destroyed:
    auditor.empty_trashcan()
    assert spline.is_alive is False


def test_degree_of_spline(auditor, spline):
    # Attribute validator prevents setting an invalid degree:
    with pytest.raises(const.DXFValueError):
        spline.dxf.degree = 0

    # But could be loaded from DXF file:
    # Hack to test entity validator for degree < 1
    spline.dxf.__dict__["degree"] = 0
    spline.audit(auditor)
    assert len(auditor.fixes) == 1
    assert auditor.fixes[0].code == AuditError.INVALID_SPLINE_DEFINITION

    # Test if invalid spline will be destroyed:
    auditor.empty_trashcan()
    assert spline.is_alive is False


def add_n_fit_points(s, count: int):
    for x in range(count):
        s.fit_points.append([x, 0, 0])


class TestFitPoints:
    @pytest.mark.parametrize("degree", [1, 2, 3, 4])
    def test_if_fit_point_count_is_valid(self, auditor, spline, degree):
        order = degree + 1
        spline.dxf.degree = degree
        add_n_fit_points(spline, required_fit_points(order, tangents=True) - 1)

        spline.audit(auditor)
        assert len(auditor.fixes) == 1
        assert (
            auditor.fixes[0].code == AuditError.INVALID_SPLINE_FIT_POINT_COUNT
        )

        # Test if invalid spline will be destroyed:
        auditor.empty_trashcan()
        assert spline.is_alive is False

    def test_remove_unused_knot_values(self, auditor, spline):
        add_n_fit_points(spline, 4)
        # Add arbitrary knot values -- have no meaning for splines defined
        # by fit points:
        spline.knots = [1, 2, 3, 4]
        spline.audit(auditor)
        assert len(auditor.fixes) == 1
        assert (
            auditor.fixes[0].code == AuditError.INVALID_SPLINE_KNOT_VALUE_COUNT
        )
        assert len(spline.knots) == 0

        # Spline is usable, test if spline will not be destroyed:
        auditor.empty_trashcan()
        assert spline.is_alive is True

    def test_remove_unused_weights(self, auditor, spline):
        add_n_fit_points(spline, 4)
        # Add arbitrary weights -- have no meaning for splines defined by fit
        # points:
        spline.weights = [1, 2, 2, 1]
        spline.audit(auditor)
        assert len(auditor.fixes) == 1
        assert auditor.fixes[0].code == AuditError.INVALID_SPLINE_WEIGHT_COUNT
        assert len(spline.weights) == 0

        # Spline is usable, test if spline will not be destroyed:
        auditor.empty_trashcan()
        assert spline.is_alive is True


def add_n_control_points(s, count: int):
    for x in range(count):
        s.control_points.append([x, 0, 0])


class TestControlPoints:
    @pytest.mark.parametrize("degree", [1, 2, 3, 4])
    def test_auditing_control_point_count(self, auditor, spline, degree):
        order = degree + 1
        spline.dxf.degree = degree
        add_n_control_points(spline, required_control_points(order) - 1)

        spline.audit(auditor)
        assert len(auditor.fixes) == 1
        assert (
            auditor.fixes[0].code
            == AuditError.INVALID_SPLINE_CONTROL_POINT_COUNT
        )

        # Test if invalid spline will be destroyed:
        auditor.empty_trashcan()
        assert spline.is_alive is False

    @pytest.mark.parametrize("degree", [1, 2, 3, 4])
    def test_auditing_knot_value_count(self, auditor, spline, degree):
        order = degree + 1
        spline.dxf.degree = degree
        add_n_control_points(spline, required_control_points(order))
        spline.knots = [1, 2, 3]

        spline.audit(auditor)
        assert len(auditor.fixes) == 1
        assert (
            auditor.fixes[0].code == AuditError.INVALID_SPLINE_KNOT_VALUE_COUNT
        )

        # Test if invalid spline will be destroyed:
        auditor.empty_trashcan()
        assert spline.is_alive is False

    @pytest.mark.parametrize("degree", [1, 2, 3, 4])
    def test_auditing_weight_count(self, auditor, spline, degree):
        order = degree + 1
        spline.dxf.degree = degree
        count = required_control_points(order)
        add_n_control_points(spline, count)
        spline.knots = uniform_knot_vector(count, order)
        spline.weights = range(count - 1)

        spline.audit(auditor)
        assert len(auditor.fixes) == 1
        assert auditor.fixes[0].code == AuditError.INVALID_SPLINE_WEIGHT_COUNT

        # Test if invalid spline will be destroyed:
        auditor.empty_trashcan()
        assert spline.is_alive is False


if __name__ == "__main__":
    pytest.main([__file__])



# Created: 21.03.2011, 2018 rewritten for pytest
# Copyright (C) 2011-2019, Manfred Moitzi
# License: MIT License
import pytest
from datetime import datetime

from ezdxf.tools.juliandate import juliandate, calendardate


class TestJulianDate:
    def test_1582_10_15(self):
        assert 2299161.0 == pytest.approx(juliandate(datetime(1582, 10, 15)))

    def test_1990_01_01(self):
        assert 2447893.0 == pytest.approx(juliandate(datetime(1990, 1, 1)))

    def test_2000_01_01(self):
        assert 2451545.0 == pytest.approx(juliandate(datetime(2000, 1, 1)))

    def test_2011_03_21(self):
        assert 2455642.75 == pytest.approx(
            juliandate(datetime(2011, 3, 21, 18, 0, 0))
        )

    def test_1999_12_31(self):
        assert 2451544.91568287 == pytest.approx(
            juliandate(datetime(1999, 12, 31, 21, 58, 35))
        )


class TestCalendarDate:
    def test_1999_12_31(self):
        check = datetime(1999, 12, 31, 21, 58, 35)
        assert calendardate(2451544.91568288) == check

    def test_2011_03_21(self):
        check = datetime(2011, 3, 21, 18, 0, 0)
        assert calendardate(2455642.75) == check



#  Copyright (c) 2020, Manfred Moitzi
#  License: MIT License

import pytest
from ezdxf.render.forms import square, translate
from ezdxf.path import Path, nesting, from_vertices

EXTERIOR = list(translate(square(10), (-5, -5)))
EXT1_PATH = from_vertices(EXTERIOR)
EXT2_PATH = from_vertices(translate(EXTERIOR, (11, 0)))

CENTER_HOLE1 = list(translate(square(8), (-4, -4)))
CH1_PATH = from_vertices(CENTER_HOLE1)

CENTER_HOLE2 = list(translate(square(6), (-3, -3)))
CH2_PATH = from_vertices(CENTER_HOLE2)

LEFT_HOLE = list(translate(square(2.1), (-3, -1)))
LH_PATH = from_vertices(LEFT_HOLE)

RIGHT_HOLE = list(translate(square(2.0), (3, -1)))
RH_PATH = from_vertices(RIGHT_HOLE)

DETECTION_DATA = [
    pytest.param(
        # Each polygon is a list of paths
        [EXT1_PATH],
        [[EXT1_PATH]],
        id="1 path",
    ),
    pytest.param(
        # returns the path sorted by area, and reversed if equal sized
        [EXT1_PATH, EXT2_PATH],
        [[EXT2_PATH], [EXT1_PATH]],
        id="2 separated paths",
    ),
    pytest.param(
        [CH1_PATH, EXT1_PATH], [[EXT1_PATH, [CH1_PATH]]], id="1 nested sub-path"
    ),
    pytest.param(
        [CH1_PATH, EXT1_PATH, CH2_PATH],
        [[EXT1_PATH, [CH1_PATH, [CH2_PATH]]]],
        id="2 nested sub-path",
    ),
    pytest.param(
        [RH_PATH, LH_PATH, EXT1_PATH],
        [[EXT1_PATH, [LH_PATH], [RH_PATH]]],
        id="2 separated sub-paths",
    ),
]


@pytest.mark.parametrize("paths,polygons", DETECTION_DATA)
def test_fast_bbox_detection(paths, polygons):
    assert nesting.make_polygon_structure(paths) == polygons


@pytest.mark.parametrize(
    "polygons,exp_ccw,exp_cw",
    [
        pytest.param(
            [[EXT1_PATH]],
            [EXT1_PATH],  # ccw paths
            [],  # cw paths
            id="1 polygon",
        ),
        pytest.param(
            [[EXT1_PATH], [EXT1_PATH]],
            [EXT1_PATH, EXT1_PATH],  # ccw paths
            [],  # cw paths
            id="2 polygons",
        ),
        pytest.param(
            [[EXT1_PATH, [CH1_PATH]]],
            [EXT1_PATH],  # ccw paths
            [CH1_PATH],  # cw paths
            id="1 polygon 1 nested sub-polygon",
        ),
        pytest.param(
            [[EXT1_PATH, [CH1_PATH, [CH1_PATH]]]],
            [EXT1_PATH, CH1_PATH],  # ccw paths
            [CH1_PATH],  # cw paths
            id="1 polygon 2 nested sub-polygons",
        ),
        pytest.param(
            [[EXT1_PATH, [CH1_PATH], [CH1_PATH]]],
            [EXT1_PATH],  # ccw paths
            [CH1_PATH, CH1_PATH],  # cw paths
            id="1 polygon 2 separated sub-polygons",
        ),
    ],
)
def test_winding_deconstruction(polygons, exp_ccw, exp_cw):
    ccw, cw = nesting.winding_deconstruction(polygons)
    assert ccw == exp_ccw
    assert cw == exp_cw


@pytest.mark.parametrize(
    "polygons,n",
    [
        pytest.param([[EXT1_PATH]], 1, id="1 polygon"),
        pytest.param([[EXT1_PATH], [EXT1_PATH]], 2, id="2 polygons"),
        pytest.param(
            [[EXT1_PATH, [CH1_PATH]]], 2, id="1 polygon 1 nested sub-polygon"
        ),
        pytest.param(
            [[EXT1_PATH, [CH1_PATH, [CH1_PATH]]]],
            3,
            id="1 polygon 2 nested sub-polygons",
        ),
        pytest.param(
            [[EXT1_PATH, [CH1_PATH], [CH1_PATH]]],
            3,
            id="1 polygon 2 separated sub-polygons",
        ),
        pytest.param(
            [[EXT1_PATH, [CH1_PATH, [CH2_PATH]], [CH1_PATH, [CH2_PATH]]]],
            5,
            id="1 polygon 2 separated nested sub-polygons",
        ),
    ],
)
def test_flatten_polygons(polygons, n):
    nlists = 0
    npaths = 0
    for path in list(nesting.flatten_polygons(polygons)):
        if isinstance(path, Path):
            npaths += 1
        elif isinstance(path, list):
            nlists += 1
        else:
            raise TypeError("?")

    assert nlists == 0
    assert npaths == n


if __name__ == "__main__":
    pytest.main([__file__])



# Copyright (c) 2018 Manfred Moitzi
# License: MIT License
import pytest
import math
from ezdxf.math import (
    bulge_radius,
    bulge_center,
    arc_to_bulge,
    bulge_3_points,
    bulge_to_arc,
    bulge_from_radius_and_chord,
    bulge_from_arc_angle,
)


def test_bulge_radius():
    assert bulge_radius((0, 0), (1, 0), 1) == 0.5
    assert bulge_radius((0, 0), (1, 0), -1) == 0.5


def test_bulge_center():
    assert bulge_center((0, 0), (1, 0), 1).isclose((0.5, 0))
    assert bulge_center((0, 0), (1, 0), -1).isclose((0.5, 0))


def test_arc_to_bulge():
    start, end, bulge = arc_to_bulge(
        center=(0.5, 0), start_angle=math.pi, end_angle=0, radius=0.5
    )
    assert start.isclose((0.0, 0.0, 0.0))
    assert end.isclose((1.0, 0.0, 0.0))
    assert math.isclose(bulge, 1.0)


def test_bulge_3_points():
    assert math.isclose(
        bulge_3_points(start_point=(0, 0), end_point=(1, 0), point=(0.5, -0.5)),
        1.0,
    )


def test_bulge_to_arc():
    center, start_angle, end_angle, radius = bulge_to_arc(
        start_point=(0, 0), end_point=(1, 0), bulge=-1
    )
    assert center.isclose((0.5, 0.0, 0.0))
    assert math.isclose(start_angle, 0, abs_tol=1e-15)
    assert math.isclose(end_angle, math.pi)
    assert radius == 0.5


class TestBulgeFromRadiusAndChord:
    def test_semi_circle_bulge(self):
        assert math.isclose(bulge_from_radius_and_chord(5.0, 10), 1.0)

    def test_half_bulge(self):
        assert math.isclose(bulge_from_radius_and_chord(6.25, 10), 0.50)

    def test_radius_of_zero(self):
        assert bulge_from_radius_and_chord(0, 10) == 0.0

    def test_too_small_radius_for_chord(self):
        assert bulge_from_radius_and_chord(1, 10) == 0.0


@pytest.mark.parametrize("angle,bulge", [
    (180, 1.0),
    (-180, -1.0),
    (106.26020471, 0.5),
    (-106.26020471, -0.5),
    (56.14497387, 0.25),
    (-56.14497387, -0.25),
])
def test_bulge_from_arc_angle(angle, bulge):
    assert math.isclose(bulge_from_arc_angle(math.radians(angle)), bulge)




# Copyright (c) 2018-2019 Manfred Moitzi
# License: MIT License
from ezdxf.lldxf.types import DXFVertex


def test_init():
    v = DXFVertex(10, (1, 2, 3))
    assert v.value == (1.0, 2.0, 3.0)


def test_clone():
    v = DXFVertex(10, (1, 2, 3))
    v2 = v.clone()
    assert v2.code == v.code
    assert v2.value == v.value


def test_dxf_tags():
    v = DXFVertex(10, (1, 2, 3))
    tags = tuple(v.dxftags())
    assert tags[0] == (10, 1.0)
    assert tags[1] == (20, 2.0)
    assert tags[2] == (30, 3.0)


def test_dxf_string():
    v = DXFVertex(10, (1, 2, 3))
    assert v.dxfstr() == " 10\n1.0\n 20\n2.0\n 30\n3.0\n"


def test_xdata_string():
    v = DXFVertex(1011, (1, 2, 3))
    assert v.dxfstr() == "1011\n1.0\n1021\n2.0\n1031\n3.0\n"



# Copyright (c) 2014, Manfred Moitzi
# License: MIT License
# This module defines all message types for the internal messaging system:
# - ezdxf.messenger
# - DXFEntity.notify()


# Notify all entities to apply pending/temporary changes.
COMMIT_PENDING_CHANGES = 1


#  Copyright (c) 2020, Manfred Moitzi
#  License: MIT License
import os
import pytest
from ezdxf import recover

BASEDIR = os.path.dirname(__file__)
DATADIR = "data"


@pytest.fixture(params=["ASCII_R12.dxf"])
def filename(request):
    filename = os.path.join(BASEDIR, DATADIR, request.param)
    if not os.path.exists(filename):
        pytest.skip(f"File {filename} not found.")
    return filename


def test_load_special_dxf_unicode_notation(filename):
    doc, auditor = recover.readfile(filename)
    layer = doc.layers.get("ΛΑΓΕΡÄÜÖ")
    assert layer.dxf.name == "ΛΑΓΕΡÄÜÖ"

    msp = doc.modelspace()
    lines = msp.query('LINE[layer=="ΛΑΓΕΡÄÜÖ"]')
    assert len(lines) == 2



# Copyright (c) 2018 Manfred Moitzi
# License: MIT License
import pytest
import ezdxf
from ezdxf.tools.standards import setup_visual_styles


@pytest.fixture(scope="module")
def doc():
    return ezdxf.new("R2007")


def test_visualstyle(doc):
    vstyle = doc.objects.new_entity(
        "VISUALSTYLE", dxfattribs={"description": "Testing", "style_type": 7}
    )
    assert vstyle.dxftype() == "VISUALSTYLE"
    assert vstyle.dxf.description == "Testing"
    assert vstyle.dxf.style_type == 7


def test_setup_standard_visual_styles(doc):
    setup_visual_styles(doc)
    styles = doc.rootdict["ACAD_VISUALSTYLE"]
    assert len(styles) == 25
    assert "Wireframe" in styles



# Copyright (c) 2018-2019 Manfred Moitzi
# License: MIT License
import pytest

from ezdxf.lldxf.extendedtags import ExtendedTags
from ezdxf.lldxf.types import is_embedded_object_marker
from ezdxf.entities.mtext import MText

MTEXT = r"""0
MTEXT
5
278
330
1F
100
AcDbEntity
8
0
100
AcDbMText
10
2762.147
20
2327.073
30
0.0
40
2.5
41
18.851
46
0.0
71
1
72
5
1
{\fArial|b0|i0|c162|p34;CHANGE;\P\P\PTEXT}
73
1
44
1.0
101
Embedded Object
70
1
10
1.0
20
0.0
30
0.0
11
2762.147
21
2327.073
31
0.0
40
18.851
41
0.0
42
15.428
43
15.042
71
2
72
1
44
18.851
45
12.5
73
0
74
0
46
0.0
"""


@pytest.fixture
def mtext_tags():
    return ExtendedTags.from_text(MTEXT)


def test_parse_embedded_object(mtext_tags):
    tags = mtext_tags
    assert tags.embedded_objects is not None
    assert len(tags.embedded_objects) == 1


def test_embedded_object_structure(mtext_tags):
    emb_obj = mtext_tags.embedded_objects[0]
    assert is_embedded_object_marker(emb_obj[0])
    assert len(emb_obj) == 15
    assert emb_obj[-1] == (46, 0.0)


def test_mtext_structure(mtext_tags):
    assert len(mtext_tags.subclasses[2]) == 10

    mtext = MText.from_text(MTEXT)
    assert mtext.dxf.handle == "278"
    assert mtext.dxf.line_spacing_factor == 1.0


def test_mtext_set_text():
    mtext = MText.from_text(MTEXT)
    mtext.text = "Hello?"
    assert mtext.text == "Hello?"
    assert mtext.dxf.line_spacing_factor == 1.0


@pytest.fixture
def two_embedded_objects():
    return ExtendedTags.from_text(
        """0
TEST
5
FFFF
101
Embedded Object
1
Text
101
Embedded Object
2
Text2
"""
    )


def test_two_embedded_objects(two_embedded_objects):
    tags = two_embedded_objects
    assert len(tags.embedded_objects) == 2

    emb_obj = tags.embedded_objects[0]
    assert is_embedded_object_marker(emb_obj[0])
    assert emb_obj[1] == (1, "Text")

    emb_obj = tags.embedded_objects[1]
    assert is_embedded_object_marker(emb_obj[0])
    assert emb_obj[1] == (2, "Text2")


def test_iter_tags(two_embedded_objects):
    tags = two_embedded_objects
    flat_tags = list(tags)
    assert len(flat_tags) == 6
    assert flat_tags[0] == (0, "TEST")
    assert flat_tags[-1] == (2, "Text2")



#  Copyright (c) 2020-2023, Manfred Moitzi
#  License: MIT License

import pytest
import ezdxf
from ezdxf.lldxf.encoding import decode_mif_to_unicode, has_mif_encoding


class TestUnicodeEncoding:
    def test_has_dxf_unicode_encoding(self):
        assert ezdxf.has_dxf_unicode(r"\U+039B") is True
        assert ezdxf.has_dxf_unicode(r"\\U+039B") is True

    def test_has_not_dxf_unicode_encoding(self):
        assert ezdxf.has_dxf_unicode(r"\U+039") is False
        assert ezdxf.has_dxf_unicode(r"\U+") is False
        assert ezdxf.has_dxf_unicode("ABC") is False
        assert ezdxf.has_dxf_unicode("") is False

    def test_decode_empty_string(self):
        assert ezdxf.decode_dxf_unicode("") == ""

    def test_decode_regular_escape_sequences(self):
        assert ezdxf.decode_dxf_unicode("\n\r\t") == "\n\r\t"

    def test_decode_regular_string_without_encoding(self):
        assert ezdxf.decode_dxf_unicode("abc") == "abc"

    def test_successive_chars(self):
        result = ezdxf.decode_dxf_unicode(r"abc\U+039B\U+0391\U+0393\U+0395\U+03A1xyz")
        assert result == r"abcΛΑΓΕΡxyz"

    def test_extra_backslash(self):
        result = ezdxf.decode_dxf_unicode(
            r"abc\U+039B\\U+0391\\U+0393\\U+0395\\U+03A1xyz"
        )
        assert result == r"abcΛ\Α\Γ\Ε\Ρxyz"

    def test_extra_digits(self):
        result = ezdxf.decode_dxf_unicode(
            r"abc\U+039B0\U+03911\U+03932\U+03953\U+03A1xyz"
        )
        assert result == "abcΛ0Α1Γ2Ε3Ρxyz"


class TestMIFEncoding:
    def test_has_mif_encoding(self):
        assert has_mif_encoding(r"\M+5D7DF\M+5CFDF\M+5BCDC") is True

    def test_has_not_mif_encoding(self):
        assert has_mif_encoding(r"M+5BCDC") is False
        assert has_mif_encoding(r"\M+5BCD") is False, "5 hex digits expected"

    def test_decode_mif_encoding(self):
        assert decode_mif_to_unicode("abc") == "abc"
        assert decode_mif_to_unicode(r"\M+5D7DF\M+5CFDF\M+5BCDC") == "走线架"
        assert decode_mif_to_unicode(r"*\M+5D7DF*") == "*走*"
        assert decode_mif_to_unicode(r"\M+5D7DF\M+5CFDFM+5BCDC") == "走线M+5BCDC"

    def test_decode_empty_string(self):
        assert decode_mif_to_unicode("") == ""

    def test_decode_regular_escape_sequences(self):
        assert decode_mif_to_unicode("\n\r\t") == "\n\r\t"

    def test_decode_regular_string_without_mif_encoding(self):
        assert decode_mif_to_unicode("abc") == "abc"


if __name__ == "__main__":
    pytest.main([__file__])



# Copyright (c) 2020, Manfred Moitzi
# License: MIT License
import pytest

from ezdxf.lldxf.validator import (
    is_in_integer_range,
    is_valid_aci_color,
    is_valid_layer_name,
    is_valid_lineweight,
    is_not_null_vector,
    is_positive,
    fix_lineweight,
    is_integer_bool,
    is_valid_one_line_text,
    fix_one_line_text,
    is_not_zero,
    is_not_negative,
    is_one_of,
    is_in_float_range,
    fit_into_float_range,
    fix_integer_bool,
    fit_into_integer_range,
    is_valid_bitmask,
    fix_bitmask,
    is_greater_or_equal_zero,
    is_handle,
    is_transparency,
    is_valid_rgb,
)
from ezdxf.entities.layer import is_valid_layer_color_index, fix_layer_color


def test_invalid_layer_name():
    assert is_valid_layer_name("Layer Layer") is True
    assert is_valid_layer_name("Layer/") is False
    assert is_valid_layer_name("Layer=") is False
    assert is_valid_layer_name("Layer;") is False
    assert is_valid_layer_name("Layer:") is False
    assert is_valid_layer_name("Layer<") is False
    assert is_valid_layer_name("Layer>") is False
    assert is_valid_layer_name("Layer`") is False
    assert is_valid_layer_name("\\Layer`") is False
    assert is_valid_layer_name('"Layer"') is False

def test_the_asterisk_in_layer_names():
    """The `*` as first letter is used by AutoCAD for internal layer names, 
    which do not show up in the layers panel. Only the first letter can be a asterisk.
    """
    assert is_valid_layer_name("*Layer") is True
    assert is_valid_layer_name("*") is False
    assert is_valid_layer_name("L*ayer") is False
    assert is_valid_layer_name("Layer*") is False

def test_allow_dxf_unicode_encoding_in_layer_names():
    assert is_valid_layer_name(r"tschüss mit \U+00FC") is True


def test_allow_dxf_mplus_encoding_in_layer_names():
    # I don't know the real name of this encoding, so I call it "mplus" encoding
    assert is_valid_layer_name(r"\M+5D7DF\M+5CFDF\M+5BCDC") is True


def test_strange_but_valid_layer_name():
    assert is_valid_layer_name("Layer|Layer") is True


def test_is_adsk_special_layer():
    assert is_valid_layer_name("*adsk_xyz") is True
    assert is_valid_layer_name("*ADSK_xyz") is True
    assert is_valid_layer_name("ADSK_xyz*") is False


def test_is_valid_lineweight():
    assert is_valid_lineweight(0) is True
    assert is_valid_lineweight(50) is True
    assert is_valid_lineweight(211) is True
    assert is_valid_lineweight(-4) is False, "is too small"
    assert is_valid_lineweight(212) is False, "is too big"
    assert is_valid_lineweight(10) is False


def test_lineweight_fixer():
    assert fix_lineweight(-4) == -1, "too small, fix as BYLAYER"
    assert fix_lineweight(212) == 211, "too big, fix as biggest lineweight"
    assert fix_lineweight(10) == 13, "invalid, fix as next valid lineweight"


def test_is_valid_aci_color():
    assert is_valid_aci_color(-1) is False
    assert is_valid_aci_color(0) is True
    assert is_valid_aci_color(257) is True
    assert is_valid_aci_color(258) is False


def test_is_in_integer_range():
    validator = is_in_integer_range(1, 10)
    assert validator(0) is False
    assert validator(1) is True
    assert validator(9) is True
    assert validator(10) is False, "exclude end value"


def test_fit_into_integer_range():
    fixer = fit_into_integer_range(0, 6)
    assert fixer(-1) == 0
    assert fixer(0) == 0
    assert fixer(5) == 5
    assert fixer(6) == 5, "exclude end value"


def test_is_in_float_range():
    validator = is_in_float_range(1, 10)
    assert validator(0) is False
    assert validator(1) is True
    assert validator(9) is True
    assert validator(10) is True, "include end value"


def test_fit_into_float_range():
    fixer = fit_into_float_range(0.25, 4.00)
    assert fixer(0.24) == 0.25
    assert fixer(0.25) == 0.25
    assert fixer(0.50) == 0.50
    assert fixer(4.00) == 4.00, "include end value"
    assert fixer(4.01) == 4.00


def test_is_not_null_vector():
    assert is_not_null_vector((0, 0, 1)) is True
    assert is_not_null_vector((0, 1, 0)) is True
    assert is_not_null_vector((1, 0, 0)) is True
    assert is_not_null_vector((0, 0, 0)) is False


def test_is_positive_value():
    assert is_positive(1) is True
    assert is_positive(0.5) is True
    assert is_positive(0) is False
    assert is_positive(0.0) is False
    assert is_positive(-1) is False


def test_is_integer_bool():
    assert is_integer_bool(0) is True
    assert is_integer_bool(1) is True
    assert is_integer_bool(2) is False
    assert is_integer_bool(-1) is False


def test_fix_integer_bool():
    assert fix_integer_bool(0) == 0
    assert fix_integer_bool(1) == 1
    assert fix_integer_bool(None) == 0
    assert fix_integer_bool("") == 0
    assert fix_integer_bool("A") == 1
    assert fix_integer_bool(2) == 1
    assert fix_integer_bool(-1) == 1


@pytest.mark.parametrize(
    "invalid_text",
    [
        "test\ntext\r",
        "test\r\ntext",
        "testtext^",
        "test\ntext^",
        "test\ntext^\r",
    ],
)
def test_is_valid_one_line_text(invalid_text):
    assert is_valid_one_line_text(invalid_text) is False


@pytest.mark.parametrize(
    "invalid_text",
    [
        "test\ntext\r",
        "test\r\ntext",
        "testtext^",
        "test\ntext^",
        "test\ntext^\r",
    ],
)
def test_fix_invalid_one_line_text(invalid_text):
    assert fix_one_line_text(invalid_text) == "testtext"


def test_is_not_negative():
    assert is_not_negative(-1) is False
    assert is_not_negative(-1e-9) is False
    assert is_not_negative(0) is True
    assert is_not_negative(1e-9) is True
    assert is_not_negative(1) is True


def test_is_not_zero():
    assert is_not_zero(-1) is True
    assert is_not_zero(-1e-9) is True
    assert is_not_zero(1e-9) is True
    assert is_not_zero(1) is True
    assert is_not_zero(0) is False
    assert is_not_zero(0.0) is False
    assert is_not_zero(1e-12) is False
    assert is_not_zero(-1e-12) is False


def test_is_one_of():
    _validator = is_one_of({1, 3, 5})
    assert _validator(0) is False
    assert _validator(2) is False
    assert _validator(4) is False
    assert _validator(1) is True
    assert _validator(3) is True
    assert _validator(5) is True


def test_is_greater_or_equal_zero():
    assert is_greater_or_equal_zero(-1) is False
    assert is_greater_or_equal_zero(0) is True
    assert is_greater_or_equal_zero(1) is True


def test_is_valid_bitmask():
    validator = is_valid_bitmask(3)
    assert validator(0) is True
    assert validator(1) is True
    assert validator(2) is True
    assert validator(3) is True
    assert validator(4) is False


def test_fix_bitmask():
    fixer = fix_bitmask(3)
    assert fixer(0) == 0
    assert fixer(1) == 1
    assert fixer(2) == 2
    assert fixer(3) == 3
    assert fixer(4) == 0
    assert fixer(5) == 1


@pytest.mark.parametrize("aci", [255, -7, -1, 1, 7, 255])
def test_is_valid_layer_color(aci):
    assert is_valid_layer_color_index(aci) is True
    assert fix_layer_color(aci) == aci


@pytest.mark.parametrize("aci", [256, 0, 256])
def test_is_not_valid_layer_color(aci):
    assert is_valid_layer_color_index(aci) is False
    assert fix_layer_color(aci) == 7


@pytest.mark.parametrize("handle", ["0", "100", "FEFE"])
def test_is_a_handle(handle):
    assert is_handle(handle) is True


@pytest.mark.parametrize("handle", [None, 0, 0x200000, "xyz"])
def test_is_not_a_handle(handle):
    assert is_handle(handle) is False


@pytest.mark.parametrize(
    "t",
    [
        0x02000000,  # 100% transparent
        0x0200007F,  # 50% transparent
        0x020000FF,  # opaque
        0x01000000,  # ByBlock
    ],
)
def test_is_transparency(t):
    assert is_transparency(t) is True


@pytest.mark.parametrize("t", [None, 0, 127, 255, 0x01000001])
def test_is_not_transparency(t):
    assert is_transparency(t) is False


@pytest.mark.parametrize("rgb", [
    (0, 0, 0),
    [0, 0, 0],
    (255, 255, 255),
    [255, 255, 255],
])
def test_is_valid_rgb(rgb):
    assert is_valid_rgb(rgb) is True


@pytest.mark.parametrize("rgb", [
    None,
    1,
    1.0,
    "1",
    (0,),
    (0, 0),
    (0, 0, 0, 0),
    (-1, 0, 0),
    (256, 0, 0),
    ("0", 0, 0),
    (0.0, 0, 0),  # no floats!
])
def test_is_not_valid_rgb(rgb):
    assert is_valid_rgb(rgb) is False


if __name__ == "__main__":
    pytest.main([__file__])



# Copyright (c) 2020-2023, Matthew Broadway
# License: MIT License
from __future__ import annotations
from typing import Iterable, Optional, Union
import math
import logging
from os import PathLike

import matplotlib.pyplot as plt
import numpy as np
from matplotlib.collections import LineCollection
from matplotlib.image import AxesImage
from matplotlib.lines import Line2D
from matplotlib.patches import PathPatch
from matplotlib.path import Path
from matplotlib.transforms import Affine2D

from ezdxf.npshapes import to_matplotlib_path
from ezdxf.addons.drawing.backend import Backend, BkPath2d, BkPoints2d, ImageData
from ezdxf.addons.drawing.properties import BackendProperties, LayoutProperties
from ezdxf.addons.drawing.type_hints import FilterFunc
from ezdxf.addons.drawing.type_hints import Color
from ezdxf.math import Vec2, Matrix44
from ezdxf.layouts import Layout
from .config import Configuration

logger = logging.getLogger("ezdxf")
# matplotlib docs: https://matplotlib.org/index.html

# line style:
# https://matplotlib.org/api/_as_gen/matplotlib.lines.Line2D.html#matplotlib.lines.Line2D.set_linestyle
# https://matplotlib.org/gallery/lines_bars_and_markers/linestyles.html

# line width:
# https://matplotlib.org/api/_as_gen/matplotlib.lines.Line2D.html#matplotlib.lines.Line2D.set_linewidth
# points unit (pt), 1pt = 1/72 inch, 1pt = 0.3527mm
POINTS = 1.0 / 0.3527  # mm -> points
CURVE4x3 = (Path.CURVE4, Path.CURVE4, Path.CURVE4)
SCATTER_POINT_SIZE = 0.1


def setup_axes(ax: plt.Axes):
    # like set_axis_off, except that the face_color can still be set
    ax.xaxis.set_visible(False)
    ax.yaxis.set_visible(False)
    for s in ax.spines.values():
        s.set_visible(False)

    ax.autoscale(False)
    ax.set_aspect("equal", "datalim")


class MatplotlibBackend(Backend):
    """Backend which uses the :mod:`Matplotlib` package for image export.

    Args:
        ax: drawing canvas as :class:`matplotlib.pyplot.Axes` object
        adjust_figure: automatically adjust the size of the parent
            :class:`matplotlib.pyplot.Figure` to display all content
    """

    def __init__(
        self,
        ax: plt.Axes,
        *,
        adjust_figure: bool = True,
    ):
        super().__init__()
        setup_axes(ax)
        self.ax = ax
        self._adjust_figure = adjust_figure
        self._current_z = 0

    def configure(self, config: Configuration) -> None:
        if config.min_lineweight is None:
            # If not set by user, use ~1 pixel
            figure = self.ax.get_figure()
            if figure:
                config = config.with_changes(min_lineweight=72.0 / figure.dpi)
        super().configure(config)
        # LinePolicy.ACCURATE is handled by the frontend since v0.18.1

    def _get_z(self) -> int:
        z = self._current_z
        self._current_z += 1
        return z

    def set_background(self, color: Color):
        self.ax.set_facecolor(color)

    def draw_point(self, pos: Vec2, properties: BackendProperties):
        """Draw a real dimensionless point."""
        color = properties.color
        self.ax.scatter(
            [pos.x],
            [pos.y],
            s=SCATTER_POINT_SIZE,
            c=color,
            zorder=self._get_z(),
        )

    def get_lineweight(self, properties: BackendProperties) -> float:
        """Set lineweight_scaling=0 to use a constant minimal lineweight."""
        assert self.config.min_lineweight is not None
        return max(
            properties.lineweight * self.config.lineweight_scaling,
            self.config.min_lineweight,
        )

    def draw_line(self, start: Vec2, end: Vec2, properties: BackendProperties):
        """Draws a single solid line, line type rendering is done by the
        frontend since v0.18.1
        """
        if start.isclose(end):
            # matplotlib draws nothing for a zero-length line:
            self.draw_point(start, properties)
        else:
            self.ax.add_line(
                Line2D(
                    (start.x, end.x),
                    (start.y, end.y),
                    linewidth=self.get_lineweight(properties),
                    color=properties.color,
                    zorder=self._get_z(),
                )
            )

    def draw_solid_lines(
        self,
        lines: Iterable[tuple[Vec2, Vec2]],
        properties: BackendProperties,
    ):
        """Fast method to draw a bunch of solid lines with the same properties."""
        color = properties.color
        lineweight = self.get_lineweight(properties)
        _lines = []
        point_x = []
        point_y = []
        z = self._get_z()
        for s, e in lines:
            if s.isclose(e):
                point_x.append(s.x)
                point_y.append(s.y)
            else:
                _lines.append(((s.x, s.y), (e.x, e.y)))

        self.ax.scatter(point_x, point_y, s=SCATTER_POINT_SIZE, c=color, zorder=z)
        self.ax.add_collection(
            LineCollection(
                _lines,
                linewidths=lineweight,
                color=color,
                zorder=z,
                capstyle="butt",
            )
        )

    def draw_path(self, path: BkPath2d, properties: BackendProperties):
        """Draw a solid line path, line type rendering is done by the
        frontend since v0.18.1
        """
        
        mpl_path = to_matplotlib_path([path])
        try:
            patch = PathPatch(
                mpl_path,
                linewidth=self.get_lineweight(properties),
                fill=False,
                color=properties.color,
                zorder=self._get_z(),
            )
        except ValueError as e:
            logger.info(f"ignored matplotlib error: {str(e)}")
        else:
            self.ax.add_patch(patch)

    def draw_filled_paths(
        self, paths: Iterable[BkPath2d], properties: BackendProperties
    ):
        linewidth = 0

        try:
            patch = PathPatch(
                to_matplotlib_path(paths, detect_holes=True),
                color=properties.color,
                linewidth=linewidth,
                fill=True,
                zorder=self._get_z(),
            )
        except ValueError as e:
            logger.info(f"ignored matplotlib error in draw_filled_paths(): {str(e)}")
        else:
            self.ax.add_patch(patch)

    def draw_filled_polygon(self, points: BkPoints2d, properties: BackendProperties):
        self.ax.fill(
            *zip(*((p.x, p.y) for p in points.vertices())),
            color=properties.color,
            linewidth=0,
            zorder=self._get_z(),
        )

    def draw_image(
        self, image_data: ImageData, properties: BackendProperties
    ) -> None:
        height, width, depth = image_data.image.shape
        assert depth == 4

        # using AxesImage directly avoids an issue with ax.imshow where the data limits
        # are updated to include the un-transformed image because the transform is applied
        # afterward. We can use a slight hack which is that the outlines of images are drawn
        # as well as the image itself, so we don't have to adjust the data limits at all here
        # as the outline will take care of that
        handle = AxesImage(self.ax, interpolation="antialiased")
        handle.set_data(np.flip(image_data.image, axis=0))
        handle.set_zorder(self._get_z())

        (
            m11,
            m12,
            m13,
            m14,
            m21,
            m22,
            m23,
            m24,
            m31,
            m32,
            m33,
            m34,
            m41,
            m42,
            m43,
            m44,
        ) = image_data.transform
        matplotlib_transform = Affine2D(
            matrix=np.array(
                [
                    [m11, m21, m41],
                    [m12, m22, m42],
                    [0, 0, 1],
                ]
            )
        )
        handle.set_transform(matplotlib_transform + self.ax.transData)
        self.ax.add_image(handle)

    def clear(self):
        self.ax.clear()

    def finalize(self):
        super().finalize()
        self.ax.autoscale(True)
        if self._adjust_figure:
            minx, maxx = self.ax.get_xlim()
            miny, maxy = self.ax.get_ylim()
            data_width, data_height = maxx - minx, maxy - miny
            if not math.isclose(data_width, 0):
                width, height = plt.figaspect(data_height / data_width)
                self.ax.get_figure().set_size_inches(width, height, forward=True)


def _get_aspect_ratio(ax: plt.Axes) -> float:
    minx, maxx = ax.get_xlim()
    miny, maxy = ax.get_ylim()
    data_width, data_height = maxx - minx, maxy - miny
    if abs(data_height) > 1e-9:
        return data_width / data_height
    return 1.0


def _get_width_height(ratio: float, width: float, height: float) -> tuple[float, float]:
    if width == 0.0 and height == 0.0:
        raise ValueError("invalid (width, height) values")
    if width == 0.0:
        width = height * ratio
    elif height == 0.0:
        height = width / ratio
    return width, height


def qsave(
    layout: Layout,
    filename: Union[str, PathLike],
    *,
    bg: Optional[Color] = None,
    fg: Optional[Color] = None,
    dpi: int = 300,
    backend: str = "agg",
    config: Optional[Configuration] = None,
    filter_func: Optional[FilterFunc] = None,
    size_inches: Optional[tuple[float, float]] = None,
) -> None:
    """Quick and simplified render export by matplotlib.

    Args:
        layout: modelspace or paperspace layout to export
        filename: export filename, file extension determines the format e.g.
            "image.png" to save in PNG format.
        bg: override default background color in hex format #RRGGBB or #RRGGBBAA,
            e.g. use bg="#FFFFFF00" to get a transparent background and a black
            foreground color (ACI=7), because a white background #FFFFFF gets a
            black foreground color or vice versa bg="#00000000" for a transparent
            (black) background and a white foreground color.
        fg: override default foreground color in hex format #RRGGBB or #RRGGBBAA,
            requires also `bg` argument. There is no explicit foreground color
            in DXF defined (also not a background color), but the ACI color 7
            has already a variable color value, black on a light background and
            white on a dark background, this argument overrides this (ACI=7)
            default color value.
        dpi: image resolution (dots per inches).
        size_inches: paper size in inch as `(width, height)` tuple, which also
            defines the size in pixels = (`width` * `dpi`) x (`height` * `dpi`).
            If `width` or `height` is 0.0 the value is calculated by the aspect
            ratio of the drawing.
        backend: the matplotlib rendering backend to use (agg, cairo, svg etc)
            (see documentation for `matplotlib.use() <https://matplotlib.org/3.1.1/api/matplotlib_configuration_api.html?highlight=matplotlib%20use#matplotlib.use>`_
            for a complete list of backends)
        config: drawing parameters
        filter_func: filter function which takes a DXFGraphic object as input
            and returns ``True`` if the entity should be drawn or ``False`` if
            the entity should be ignored

    """
    from .properties import RenderContext
    from .frontend import Frontend
    import matplotlib

    # Set the backend to prevent warnings about GUIs being opened from a thread
    # other than the main thread.
    old_backend = matplotlib.get_backend()
    matplotlib.use(backend)
    if config is None:
        config = Configuration()

    try:
        fig: plt.Figure = plt.figure(dpi=dpi)
        ax: plt.Axes = fig.add_axes((0, 0, 1, 1))
        ctx = RenderContext(layout.doc)
        layout_properties = LayoutProperties.from_layout(layout)
        if bg is not None:
            layout_properties.set_colors(bg, fg)
        out = MatplotlibBackend(ax)
        Frontend(ctx, out, config).draw_layout(
            layout,
            finalize=True,
            filter_func=filter_func,
            layout_properties=layout_properties,
        )
        # transparent=True sets the axes color to fully transparent
        # facecolor sets the figure color
        # (semi-)transparent axes colors do not produce transparent outputs
        # but (semi-)transparent figure colors do.
        if size_inches is not None:
            ratio = _get_aspect_ratio(ax)
            w, h = _get_width_height(ratio, size_inches[0], size_inches[1])
            fig.set_size_inches(w, h, True)
        fig.savefig(filename, dpi=dpi, facecolor=ax.get_facecolor(), transparent=True)
        plt.close(fig)
    finally:
        matplotlib.use(old_backend)
